"""File containing the copulAX implementation of the generalized hyperbolic distribution."""
import jax.numpy as jnp
from jax import lax, custom_vjp, random
from jax._src.typing import ArrayLike, Array
from copy import deepcopy

from copulax._src._distributions import Univariate
from copulax._src.univariate._utils import _univariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.typing import Scalar
from copulax._src.univariate._cdf import _cdf, cdf_bwd, _cdf_fwd
from copulax.special import kv
from copulax._src.univariate._rvs import mean_variance_sampling
from copulax._src.univariate._mean_variance import mean_variance_stats, mean_variance_ldmle_params
from copulax._src.univariate.gig import gig
from copulax._src.optimize import projected_gradient


class GHBase(Univariate):
    @classmethod
    def _params_dict(cls, lamb: Scalar, chi: Scalar, psi: Scalar, mu: Scalar, sigma: Scalar, gamma: Scalar) -> dict:
        r"""Convert parameters to a dictionary."""
        d: dict = {"lambda": lamb, "chi": chi, "psi": psi, 
                   "mu": mu, "sigma": sigma, "gamma": gamma}
        return cls._args_transform(d)
                
    @staticmethod
    def _params_to_tuple(params: dict) -> tuple:
        params = GHBase._args_transform(params)
        return (params["lambda"], params["chi"], params["psi"], 
                params["mu"], params["sigma"], params["gamma"])
    
    @staticmethod
    def _params_to_array(params: dict) -> Array:
        return jnp.asarray(GHBase._params_to_tuple(params)).flatten()
    
    @classmethod
    def _support(cls, *args, **kwargs) -> tuple:
        return -jnp.inf, jnp.inf
    
    def example_params(self, *args, **kwargs) -> dict:
        r"""Example parameters for the generalized hyperbolic 
        distribution.
        
        This is a six parameter family, with the generalized hyperbolic
        being defined by location `mu`, dispersion `sigma` and skewness 
        `gamma` in addition to `lambda`, `chi` and `psi` shape parameters. 
        Here, we adopt the parameterization used by McNeil et al. (2005).
        """
        return self._params_dict(lamb=0.0, chi=1.0, psi=1.0, 
                                 mu=0.0, sigma=1.0, gamma=0.0)
    
    @staticmethod
    def _stable_logpdf(stability: Scalar, x: ArrayLike, params: dict) -> Array:
        lamb, chi, psi, mu, sigma, gamma = GHBase._params_to_tuple(params)
        x, xshape = _univariate_input(x)

        r: float = lax.sqrt(lax.mul(chi, psi))
        s: float = 0.5 - lamb
        h: float = lax.add(psi, lax.pow(lax.div(gamma, sigma), 2))
        g = lax.div(lax.sub(x, mu), lax.pow(sigma, 2))

        m = lax.sqrt(lax.mul(lax.add(chi, lax.mul(g, lax.sub(x, mu))), h))

        T = lax.add(lax.log(kv(-s, m) + stability), lax.mul(g, gamma))
        B = lax.mul(lax.log(m + stability), s)

        cT = lax.add(lax.mul(lamb, lax.log((psi / (r + stability)) + stability)), lax.mul(lax.log(h), s)) 
        cB = lax.add(lax.add(lax.log(sigma), lax.log(lax.sqrt(2*jnp.pi))), lax.log(kv(lamb, r) + stability))

        c = lax.sub(cT, cB)
        logpdf: jnp.ndarray = lax.add(lax.sub(T, B), c)
        return logpdf.reshape(xshape)
    
    @staticmethod
    def logpdf(x: ArrayLike, params: dict) -> Array:
        return GHBase._stable_logpdf(stability=0.0, x=x, params=params)
    
    @staticmethod
    def pdf(x: ArrayLike, params: dict) -> Array:
        return lax.exp(GHBase.logpdf(x=x, params=params))
    
    # # ppf
    # def _get_x0(self, params: dict) -> Scalar:
    #     return self.stats(params=params)["mean"]
    
    # sampling
    def rvs(self, size: tuple | Scalar, params: dict, key: Array = DEFAULT_RANDOM_KEY) -> Array:
        lamb, chi, psi, mu, sigma, gamma = self._params_to_tuple(params)

        key1, key2 = random.split(key)
        W = gig.rvs(key=key1, size=size, params={'lambda': lamb, 'chi': chi, 'psi': psi})
        return mean_variance_sampling(key=key2, W=W, shape=size, mu=mu, sigma=sigma, gamma=gamma)

    # stats
    def _get_w_stats(self, lamb: Scalar, chi: Scalar, psi: Scalar) -> dict:
        return gig.stats(params={'lambda': lamb, 'chi': chi, 'psi': psi})

    def stats(self, params: dict) -> dict:
        lamb, chi, psi, mu, sigma, gamma = self._params_to_tuple(params) 
        gig_stats: dict = self._get_w_stats(lamb=lamb, chi=chi, psi=psi)
        return self._scalar_transform(mean_variance_stats(w_stats=gig_stats, mu=mu, sigma=sigma, gamma=gamma))

    # fitting
    def _fit_mle(self, x: jnp.ndarray, lr: float, maxiter: int) -> dict:
        eps: float = 1e-8
        constraints: tuple = (jnp.array([[-jnp.inf, eps, eps, -jnp.inf, eps, -jnp.inf]]).T, 
                          jnp.array([[jnp.inf, jnp.inf, jnp.inf, jnp.inf, jnp.inf, jnp.inf]]).T)
        
        projection_options: dict = {'hyperparams': constraints}

        key1, key = random.split(DEFAULT_RANDOM_KEY)
        key2, key3 = random.split(key)
        params0: jnp.ndarray = jnp.array([
            random.normal(key1, ()), 
            random.uniform(key2, (), minval=eps), 
            random.uniform(key3, (), minval=eps),  
            x.mean(), 
            x.std(), 
            0.0])
        
        res: dict = projected_gradient(
            f=self._mle_objective, x0=params0, projection_method='projection_box', 
            projection_options=projection_options, x=x, lr=lr, maxiter=maxiter)
        lamb, chi, psi, mu, sigma, gamma = res['x']
        return GHBase._params_dict(lamb=lamb, chi=chi, psi=psi, mu=mu, sigma=sigma, gamma=gamma)
    
    def _ldmle_objective(self, params: jnp.ndarray, x: jnp.ndarray, sample_mean: Scalar, sample_variance: Scalar) -> Scalar:
        lamb, chi, psi, gamma = params
        gig_stats: dict = self._get_w_stats(lamb=lamb, chi=chi, psi=psi)
        mu, sigma = mean_variance_ldmle_params(stats=gig_stats, gamma=gamma, sample_mean=sample_mean, sample_variance=sample_variance)
        return self._mle_objective(params_arr=jnp.array([lamb, chi, psi, mu, sigma, gamma]), x=x)
    
    def _fit_ldmle(self, x: jnp.ndarray, lr: float, maxiter: int) -> dict:
        eps = 1e-8
        constraints: tuple = (jnp.array([[-jnp.inf, eps, eps, -jnp.inf]]).T, 
                            jnp.array([[jnp.inf, jnp.inf, jnp.inf, jnp.inf]]).T)
        
        key1, key = random.split(DEFAULT_RANDOM_KEY)
        key2, key3 = random.split(key)
        params0: jnp.ndarray = jnp.array([random.normal(key1, ()), 
                                        random.uniform(key2, (), minval=eps), 
                                        random.uniform(key3, (), minval=eps),  
                                        0.0])

        projection_options: dict = {'hyperparams': constraints}

        sample_mean, sample_variance = x.mean(), x.var()
        res = projected_gradient(
            f=self._ldmle_objective, x0=params0, x=x, lr=lr, maxiter=maxiter,
            projection_method='projection_box', projection_options=projection_options, 
            sample_mean=sample_mean, sample_variance=sample_variance)
        lamb, chi, psi, gamma = res['x']
        gig_stats: dict = self._get_w_stats(lamb=lamb, chi=chi, psi=psi)
        mu, sigma = mean_variance_ldmle_params(stats=gig_stats, gamma=gamma, sample_mean=sample_mean, sample_variance=sample_variance)
        return self._params_dict(lamb=lamb, chi=chi, psi=psi, mu=mu, sigma=sigma, gamma=gamma)
    
    def fit(self, x: ArrayLike, method: str = 'LDMLE', lr: float = 0.1, 
            maxiter: int = 100) -> dict:
        r"""Fit the distribution to the input data.

        Note:
            If you intend to jit wrap this function, ensure that 'method' is a 
            static argument.
        
        Args:
            x (ArrayLike): The input data to fit the distribution to.
            method (str): The fitting method to use.  Options are 
                'MLE' for maximum likelihood estimation, 
                and 'LDMLE' for low-dimensional maximum likelihood 
                estimation. Defaults to 'LDMLE'. 
            lr (float): Learning rate for optimization.
            maxiter (int): Maximum number of iterations for optimization.
        
        Returns:
            dict: The fitted distribution parameters.
        """ 
        x = _univariate_input(x)[0]
        if method == 'MLE':
            return self._fit_mle(x, lr=lr, maxiter=maxiter)
        else:
            return self._fit_ldmle(x, lr=lr, maxiter=maxiter)
        
    # cdf
    @staticmethod
    def _params_from_array(params_arr: jnp.ndarray, *args, **kwargs) -> dict:
        lamb, chi, psi, mu, sigma, gamma = params_arr
        return GHBase._params_dict(lamb=lamb, chi=chi, psi=psi, mu=mu, sigma=sigma, gamma=gamma)

    @staticmethod
    def _pdf_for_cdf(x: ArrayLike, *params_tuple) -> Array:
        params_array: jnp.ndarray = jnp.asarray(params_tuple).flatten()
        params: dict = GHBase._params_from_array(params_array)
        return GHBase.pdf(x=x, params=params)
    
def _vjp_cdf(x: ArrayLike, params: dict) -> Array:
    params = GHBase._args_transform(params)
    return _cdf(dist=GHBase, x=x, params=params)
    

_vjp_cdf_copy = deepcopy(_vjp_cdf)
_vjp_cdf = custom_vjp(_vjp_cdf)


def cdf_fwd(x: ArrayLike, params: dict) -> tuple[Array, tuple]:
    params = GHBase._args_transform(params)
    return _cdf_fwd(dist=GHBase, cdf_func=_vjp_cdf_copy, x=x, params=params)

_vjp_cdf.defvjp(cdf_fwd, cdf_bwd)


class GH(GHBase):
    r"""The generalized hyperbolic distribution. This is a flexible, 
    continuous 6-parameter family of distributions that can model a variety 
    of data behaviors, including heavy tails and skewness. It contains 
    a number of popular distributions as special cases, including the
    normal, student-t, hyperbolic, laplace, and skewed-T distributions.

    We adopt the parameterization used by McNeil et al. (2005):
    
    .. math::

        f(x|\mu, \sigma, \chi, \psi, \gamma, \lambda) \propto e^{\frac{(x-\mu)\gamma}{\sigma^2}} \frac{K_{\lambda - 0.5}(\sqrt{(\chi + (\frac{x-\mu}{\sigma})^2)(\psi + (\frac{\gamma}{\sigma})^2 })}{(\sqrt{(\chi + (\frac{x-\mu}{\sigma})^2)(\psi + (\frac{\gamma}{\sigma})^2 })^{0.5-\lambda}}

    where :math:`K_{\lambda}` is the modified Bessel function of the second 
    kind, :math:`\mu` is the location parameter, :math:`\sigma` is the scale,
    :math: `\gamma` is the skewness and :math:`\lambda`, :math:`\chi` and 
    :math:`\psi` relate to the shape of the distribution.
    """
    def cdf(self, x: ArrayLike, params: dict) -> Array:
        return _vjp_cdf(x=x, params=params)
    

gh = GH("GH")


"""Module containing base classes for all distributions to inherit from.
"""
from dataclasses import dataclass, field
from jax.tree_util import register_pytree_node
from abc import abstractmethod
from jax._src.typing import ArrayLike, Array
import jax.numpy as jnp
from jax import lax, jit, random
import matplotlib.pyplot as plt
from typing import Iterable


from copulax._src.typing import Scalar
from copulax._src.univariate._ppf import _ppf
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.univariate._rvs import inverse_transform_sampling
from copulax._src.typing import Scalar
from copulax._src.multivariate._utils import _multivariate_input
from copulax._src.multivariate._shape import cov, corr
from copulax._src.optimize import projected_gradient
from copulax._src.univariate._utils import _univariate_input

###############################################################################
# Record of implemented distributions
###############################################################################
@dataclass(frozen=True)
class DistMap:
    continuous_names: tuple # tuple of continuous univariate distribution names
    discrete_names: tuple # tuple of discrete univariate distribution names
    mvt_names: tuple
    copula_names: tuple

    univariate_names: tuple = field(init=False)

    ids: tuple = field(init=False)  # tuple of all distribution ids
    names: tuple = field(init=False)  # tuple of all distribution names
    dtypes: tuple = field(init=False)  # tuple of all distribution data types
    dist_types: tuple = field(init=False)  # tuple of all distribution types
    object_names: tuple = field(init=False)  # tuple of all distribution object names

    id_map: dict = field(init=False)  # dict mapping distribution ids to names and dist_type
    name_map: dict = field(init=False)  # dict mapping distribution names to ids and dist_type
    object_name_map: dict = field(init=False)  # dict mapping distribution object names to ids and dist_type

    def __post_init__(self):
        univariate_names: tuple = self.continuous_names + self.discrete_names

        id_map: dict = {}
        name_map: dict = {}
        object_name_map: dict = {}

        d: dict = {'univariate': univariate_names, 
                   'multivariate': self.mvt_names, 
                   'copula': self.copula_names}
        start: int = 0
        for s, names in d.items():
            for i, name in enumerate(names, start=start):
                dtype: str = 'discrete' if name in discrete_names else 'continuous'
                object_name: str = name.replace('-', '_').lower()
                entry: dict = {'id': i, 
                               'name': name, 
                               'object_name': object_name, 
                               'dtype': dtype, 
                               'dist_type': s}
                id_map[i] = entry
                name_map[name] = entry
                object_name_map[object_name] = entry
            start = i + 1

        super().__setattr__('univariate_names', univariate_names)
        super().__setattr__('ids', tuple(id_map.keys()))
        super().__setattr__('names', tuple(name_map.keys()))
        super().__setattr__('object_names', tuple(object_name_map.keys()))
        super().__setattr__('dtypes', ('continuous', 'discrete'))
        super().__setattr__('dist_types', tuple(d.keys()))
        super().__setattr__('id_map', id_map)
        super().__setattr__('name_map', name_map)
        super().__setattr__('object_name_map', object_name_map)



continuous_names: tuple = ("Uniform", "Normal", "LogNormal", "Student-T", 
                           "Gamma", "Skewed-T", "GIG", "GH", "IG")
discrete_names: tuple = ()
mvt_names: tuple = ("Mvt-Normal", "Mvt-Student-T", "Mvt-GH", "Mvt-Skewed-T",)
copula_names: tuple = ("Gaussian-Copula", "Student-T-Copula", "GH-Copula", "Skewed-T-Copula",)

dist_map = DistMap(continuous_names=continuous_names, 
                   discrete_names=discrete_names, mvt_names=mvt_names, 
                   copula_names=copula_names)


###############################################################################
# Distribution PyTree / base class
###############################################################################
class Distribution:
    r"""Base class for all implemented copulAX distributions."""
    def __init__(self, name: str):
        self._id: int = dist_map.name_map[name]['id']

    def __str__(self):
        return self.name
    
    def __repr__(self):
        return self.name

    @classmethod
    def tree_unflatten(cls, aux_data: tuple, values: tuple, **init_kwargs):
        """
        Args:
            aux_data:
                Data that will be treated as constant through JAX 
                operations.
            values:
                A JAX PyTree of values from which the object is 
                constructed.
        Returns:
            A constructed object.
        """
        id_: int = aux_data[0]
        return cls(dist_map.id_map[id_]['name'], **init_kwargs)
    
    def tree_flatten(self):
        """
        Returns:
            values: A JAX PyTree of values representing the object.
            aux_data:
                Data that will be treated as constant through JAX 
                operations.
        """
        children = ()  # arrays and pytrees
        aux_data = (self._id,)  # static, hashable data
        return children, aux_data
    
    def __init_subclass__(cls, **kwargs):
        # https://github.com/jax-ml/jax/issues/2916
        super().__init_subclass__(**kwargs)
        register_pytree_node(cls, cls.tree_flatten, cls.tree_unflatten)

    @property
    def name(self) -> str:
        """The name of the distribution.
        
        Returns:
            str: The name of the distribution.
        """
        return dist_map.id_map[self._id]['name']
    
    @property
    def dist_type(self) -> str:
        """The type of copulAX distribution.
        
        Returns:
            str: The type of the distribution.
        """
        return dist_map.id_map[self._id]['dist_type']
    
    @property
    def dtype(self) -> str:
        """The data type of the distribution.
        
        Returns:
            str: The data type of the distribution.
        """
        return dist_map.id_map[self._id]['dtype']
    
    @staticmethod
    def _scalar_transform(params: dict) -> dict:
        return {key: jnp.asarray(value, dtype=float).reshape(()) 
                for key, value in params.items()}
    
    @abstractmethod
    def _args_transform(self, params: dict) -> dict:
        r"""Transforms the input arguments to the correct dtype and 
        shape.
        """

    def _params_from_array(self, params_arr: jnp.ndarray, *args, **kwargs) -> dict:
        r"""Returns a dictionary from an array / iterable of params"""
        return self._params_dict(*params_arr)

    @abstractmethod
    def _params_to_tuple(self, params: dict) -> tuple:
        r"""Returns a tuple of params from a dictionary. 
        Reduces code when extracting params."""
        pass
    
    @abstractmethod
    def fit(self, x: ArrayLike, *args, **kwargs):
        r"""Fit the distribution to the input data.
        
        Args:
            x (ArrayLike): The input data to fit the distribution to.
            kwargs: Additional keyword arguments to pass to the fit 
                method.
        
        Returns:
            dict: The fitted distribution parameters.
        """

    def _params_to_array(self, params) -> Array:
        r"""Returns a flattened array of params from a dictionary. 
        Reduces code when extracting params."""
        return jnp.asarray(self._params_to_tuple(params)).flatten()
    
    @abstractmethod
    def rvs(self, size, params: dict, key: Array = DEFAULT_RANDOM_KEY, 
            *args, **kwargs) -> Array:
        r"""Generate random variates from the distribution.
        
        Args:
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.
        
        Returns:
            jnp.ndarray: The generated random variates.
        """

    def sample(self, size, params: dict, key: Array = DEFAULT_RANDOM_KEY, 
               *args, **kwargs) -> Array:
        """Alias for the rvs method."""
        return self.rvs(size=size, params=params, key=key, *args, **kwargs)

    @abstractmethod
    def fit(self, x: ArrayLike, *args, **kwargs):
        """Fit the distribution to the input data.
        
        Args:
            x (ArrayLike): The input data to fit the distribution to.
            kwargs: Additional keyword arguments to pass to the fit 
                method.
        
        Returns:
            dict: The fitted distribution parameters.
        """

    # fitting
    def _stable_logpdf(self, stability: Scalar, x: ArrayLike, 
                       params: dict) -> Array:
        r"""Stable log-pdf function for distribution fitting. 
        Utilises a stability term to help prevent the logpdf from 
        blowing to inf / nan during numerical optimisation, typically 
        resulting from log and 1 / x functions.

        Args:
            stability (Scalar): A stability parameter for the distribution.
            x (ArrayLike): The input data to evaluate the stable log-pdf.
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.  

        Returns:
            Array: The stable log-pdf values.
        """

        pass

    @abstractmethod
    def logpdf(self, x: ArrayLike, params: dict) -> Array:
        r"""The log-probability density function (pdf) of the 
        distribution.

        Args:
            x (ArrayLike): The input at which to evaluate the log-pdf.
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.  

        Returns:
            Array: The log-pdf values.
        """
        return self._stable_logpdf(stability=0.0, x=x, params=params)

    @abstractmethod
    def pdf(self, x: ArrayLike, params: dict) -> Array:
        r"""The probability density function (pdf) of the distribution.

        Args:
            x (ArrayLike): The input at which to evaluate the pdf.
            params (dict): Parameters describing the distribution. See
                    the specific distribution class or the 'example_params' 
                    method for details.
        
        Returns:
            Array: The pdf values.
        """
        return jnp.exp(self.logpdf(x=x, params=params))
    
    # stats
    @abstractmethod
    def stats(self, params: dict) -> dict:
        r"""Distribution statistics for the distribution.

        Args:
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.

        Returns:
            stats (dict): A dictionary containing the distribution 
            statistics.
        """
        return {}
    
    # metrics
    @abstractmethod
    def loglikelihood(self, x: ArrayLike, params: dict) -> Scalar:
        r"""Log-likelihood of the distribution given the data.

        Args:
            x (ArrayLike): The input data to evaluate the 
            log-likelihood.
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.  
        
        Returns:
            loglikelihood (Scalar): The log-likelihood value.
        """
        return self.logpdf(x=x, params=params).sum()

    @abstractmethod
    def aic(self, k: int, x: ArrayLike, params: dict) -> Scalar:
        r"""Akaike Information Criterion (AIC) of the distribution 
        given the data. Can be used as a crude metric for model 
        selection, by minimising.

        Args:
            x (ArrayLike): The input data to evaluate the AIC.
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.  

        Returns:
            aic (Scalar): The AIC value.
        """
        return 2 * k - 2 * self.loglikelihood(x=x, params=params)
    
    @abstractmethod
    def bic(self, k: int, n: int, x: ArrayLike, params: dict) -> Scalar:
        r"""Bayesian Information Criterion (BIC) of the distribution 
        given the data. Can be used as a crude metric for model 
        selection, by minimising.

        Args:
            x (ArrayLike): The input data to evaluate the BIC.
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.  

        Returns:
            bic (Scalar): The BIC value.
        """
        return k * jnp.log(n) - 2 * self.loglikelihood(x=x, params=params)
    
    @abstractmethod
    def example_params(self, *args, **kwargs) -> dict:
        r"""Returns example parameters for the distribution.

        Returns:
            dict: A dictionary containing example distribution 
            parameters.
        """
        pass


###############################################################################
# Univariate Base Class
###############################################################################
class Univariate(Distribution):
    r"""Base class for univariate distributions."""
    @staticmethod
    def _args_transform(params: dict) -> dict:
        return Distribution._scalar_transform(params)

    @abstractmethod
    def _support(self, *args, **kwargs) -> tuple:
        """needed for ppf."""
        pass
    
    @classmethod
    def support(cls, *args, **kwargs) -> Array:
        r"""The support of the distribution is the subset of x for which 
        the pdf is non-zero. 
        
        Returns:
            Array: Flattened array containing the support of the 
            distribution.
        """
        return jnp.array(cls._support(*args, **kwargs)).flatten()
    
    @abstractmethod
    def logcdf(self, x: ArrayLike, params: dict) -> Array:
        r"""The log-cumulative distribution function of the 
        distribution.

        Args:
            x (ArrayLike): The input at which to evaluate the log-cdf.
            params (dict): Parameters describing the distribution. See
                the specific distribution class or the 'example_params'
                method for details.

        Returns:
            Array: The log-cdf values.
        """
        return jnp.log(self.cdf(x=x, params=params))
    
    # cdf
    @classmethod
    def _params_from_array(cls, params_arr, *args, **kwargs):
        return cls._params_dict(*params_arr)

    @classmethod
    def _pdf_for_cdf(cls, x: ArrayLike, *params_tuple) -> Array:
        params_array: jnp.ndarray = jnp.asarray(params_tuple).flatten()
        params: dict = cls._params_from_array(params_array)
        return cls.pdf(x=x, params=params)
    
    @abstractmethod
    def cdf(self, x: ArrayLike, params: dict) -> Array:
        r"""Cumulative distribution function of the distribution.

        Args:
            x (ArrayLike): The input at which to evaluate the cdf.
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.

        Returns:
            Array: The cdf values.
        """

    # ppf
    # @abstractmethod
    # def _get_x0(self, params: dict) -> Scalar:
    #     """Returns the initial guess for the ppf function."""
    #     pass 

    def _ppf(self, q: ArrayLike, params: dict, cubic: bool, 
             num_points: int, maxiter: int) -> Array:
        return _ppf(dist=self, q=q, params=params, cubic=cubic, 
                    num_points=num_points, maxiter=maxiter)

    def ppf(self, q: ArrayLike, params: dict, cubic: bool = False, 
            num_points: int = 100, maxiter: int = 50) -> Array:
        r"""Percent point function (inverse of the CDF) of the 
        distribution.

        Note:
            If you intend to jit wrap this function, ensure that 'cubic' 
            is a static argument.


        Args:
            q (ArrayLike): The quantile values. at which to evaluate the 
            ppf.
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.
            cubic (bool): Whether to use a cubic spline approximation 
                of the ppf function for faster computation. This can 
                also improve gradient estimates.
            num_points (int): The number of points to use for the cubic 
                spline approximation when approx is True.
            maxiter (int): The maximum number of iterations to use when 
                solving for the ppf function via brents method.

        Returns:
            Array: The inverse CDF values.
        """
        q, qshape = _univariate_input(q)
        if cubic: 
            # approximating even if an analytical / more efficient solution exists
            x: jnp.ndarray = _ppf(dist=self, q=q, params=params, 
                                  cubic=True, num_points=num_points, maxiter=maxiter)
        else: 
            x: jnp.ndarray = self._ppf(q=q, params=params, cubic=False, 
                                       num_points=num_points, maxiter=maxiter)
        return x.reshape(qshape)
    
    @abstractmethod
    def inverse_cdf(self, q: ArrayLike, params: dict, cubic: bool = False, 
            num_points: int = 100, maxiter: int = 50) -> Array:
        r"""Percent point function (inverse of the CDF) of the 
        distribution.

        Note:
            If you intend to jit wrap this function, ensure that 'cubic' 
            is a static argument.

        Args:
            q (ArrayLike): The quantile values. at which to evaluate the 
            ppf.
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.
            cubic (bool): Whether to use a cubic spline approximation 
                of the ppf function for faster computation. This can 
                also improve gradient estimates.
            num_points (int): The number of points to use for the cubic 
                spline approximation when approx is True.
            lr (float): The learning rate to use when numerically 
                solving for the ppf function via ADAM based gradient 
                descent.
            maxiter (int): The maximum number of iterations to use when 
                solving for the ppf function via ADAM based gradient 
                descent.

        Returns:
            Array: The inverse CDF values.
        """
        return self.ppf(q=q, params=params, cubic=cubic, num_points=num_points, 
                        maxiter=maxiter)

    # sampling
    @abstractmethod
    def rvs(self, size: Scalar | tuple, params: dict,
            key: Array = DEFAULT_RANDOM_KEY) -> Array:
        r"""Generates random samples from the distribution.

        Note:
            If you intend to jit wrap this function, ensure that 'size' 
            is a static argument.

        Args:
            size (tuple | Scalar): The size / shape of the generated 
                output array of random numbers. If a scalar is provided, 
                the output array will have shape (size,), otherwise it will 
                match the shape specified by this tuple.
            params (dict): Parameters describing the distribution. See
                the specific distribution class or the 'example_params'
                method for details.
            key (Array): The Key for random number generation.
        """
        return inverse_transform_sampling(ppf_func=self.ppf, shape=size, 
                                          params=params, key=key)        
    
    @abstractmethod
    def sample(self, size: Scalar | tuple, params: dict, 
               key: Array = DEFAULT_RANDOM_KEY) -> Array:
        r"""Generates random samples from the distribution.

        Note:
            If you intend to jit wrap this function, ensure that 'size' 
            is a static argument.

        Args:
            size (tuple | Scalar): The size / shape of the generated 
                output array of random numbers. If a scalar is provided, 
                the output array will have shape (size,), otherwise it will 
                match the shape specified by this tuple.
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.
            key (Array): The Key for random number generation.
        """
        return self.rvs(size=size, params=params, key=key)
    
    # fitting
    def _mle_objective(self, params_arr: jnp.ndarray, x: jnp.ndarray, 
                       *args, **kwargs) -> Scalar:
        r"""Negative log-likelihood of the distribution given the data.

        Args:
            x (ArrayLike): The input data to evaluate the 
                negative log-likelihood.

        Returns:
            mle_objective (float): The negative log-likelihood value.
        """
        params: dict = self._params_from_array(params_arr, *args, **kwargs)
        return -self._stable_logpdf(stability=1e-30, x=x, params=params).sum()
    
    # metrics
    @abstractmethod
    def aic(self, x: ArrayLike, params: dict) -> float:
        k: int = len(params)
        return super().aic(k=k, x=x, params=params)
    
    @abstractmethod
    def bic(self, x: ArrayLike, params: dict) -> float:
        k: int = len(params)
        n: int  = x.size
        return super().bic(k=k, n=n, x=x, params=params)
    
    def plot(self, params: dict, sample: jnp.ndarray = None, domain: tuple = None, bins: int=50, num_points: int = 100, figsize: tuple = (16, 8), grid: bool = True, show: bool = True, ppf_options: dict = None):
        r"""Plots the pdf, cdf and ppf of the distribution.

        Note:
            Not jittable.

        Args:
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.
            sample (jnp.ndarray): Sample data to plot alongside the 
                distribution, allowing for a visual goodness of fit via 
                QQ plots. Must be a univariate sample if provided. 
            domain (tuple): The domain of the distribution to plot over.
                Must be a tuple of the form (min, max). If None, the ppf 
                function will be used to generate the domain.
            bins (int): The number of bins to use for the histogram of
                the sample data, if provided.
            num_points (int): The number of points to use for plotting.
            figsize (tuple): The size of the figure for the plot.
            grid (bool): Whether to display a grid on the plot.
            show (bool): Whether to automatically show the plot by 
                internally calling plt.show().
            ppf_options (dict): Options for the ppf function. This is 
                kwargs style dictionary with keyword arguments for the
                keys paired with their assigned values. See the ppf 
                method function documentation for more details.
   Returns:
            None
        """
        params: dict = self._args_transform(params=params)

        # getting pdf and cdf domain
        if ppf_options is None:
            ppf_options = {}
        jitted_ppf = jit(self.ppf, static_argnames=('cubic', 'maxiter'))
        delta: float = 1e-2
        if domain is None:
            support = self.support(params=params)

            # lower bound
            min_val, eps = support[0], 0.0
            while not jnp.isfinite(min_val):
                eps += delta
                min_val = jitted_ppf(q=jnp.array(eps), params=params, **ppf_options)

            # upper bound
            max_val, eps = support[1], 0.0
            while not jnp.isfinite(max_val):
                eps += delta
                max_val = jitted_ppf(q=jnp.array(1 - eps), params=params, **ppf_options)
        else:
            if (not isinstance(domain, Iterable)) or len(domain) != 2:
                raise ValueError("Domain must be a tuple of the form (min, max).")
            if not all(isinstance(i, Scalar) for i in domain):
                raise ValueError("Domain elements must be scalar values")
            if domain[0] >= domain[1]:
                raise ValueError("Domain must be a tuple of the form (min, max).")
            
            min_val, max_val = jnp.asarray(domain).flatten()

        x: Array = jnp.linspace(min_val, max_val, num_points)
        q: Array = jnp.linspace(delta, 1-delta, num_points)

        # pdf, cdf, ppf and rvs values
        pdf_vals: Array = self.pdf(x=x, params=params)
        cdf_vals: Array = self.cdf(x=x, params=params)
        ppf_vals: Array = jitted_ppf(q=q, params=params, **ppf_options)

        # plotting setup
        values: tuple = [pdf_vals, cdf_vals, ppf_vals]
        domains: tuple = [x, x, q]
        titles: tuple = ("PDF", "CDF", "Inverse CDF", "QQ-Plot")
        xlabels: tuple = ("x", "x", "P(X <= q)", "Theoretical Quantiles")
        ylabels: tuple = ("PDF", "P(X <= x)", "q", "Empirical Quantiles")
        dtype = float if self.dtype == 'continuous' else int
        printable_params: str = str({k: round(dtype(v), 2) for k, v in params.items()})
        name_with_params: str = f"{self.name}({printable_params})"

        # qq-plot
        if sample is not None:
            sample: Array = _univariate_input(sample)[0]
            sorted_sample: Array = sample.flatten().sort()
            N: int = sample.size
            empirical_sample_cdf = jnp.array([(sorted_sample <= xi).sum() / N 
                                              for xi in sorted_sample])
            theoretical_sample_cdf = self.cdf(x=sorted_sample, params=params)
            domains.append(theoretical_sample_cdf)
            values.append(empirical_sample_cdf)
        
        # plotting
        num_subplots = len(values)
        fig, ax = plt.subplots(1, num_subplots, figsize=figsize)
        fig.suptitle(name_with_params, fontsize=16)
        for i in range(num_subplots):
            if i == 0 and num_subplots == 4:
                ax[i].hist(sample, bins=bins, density=True, color='blue', label='Sample', zorder=0)
                ax[i].set_xlim(min_val, max_val)
            elif i == 3:
                ax[i].plot(cdf_vals, cdf_vals, color='blue', zorder=0, label='y=x')

            # plotting distribution
            if i < 3:
                ax[i].plot(domains[i], values[i], label=name_with_params, color='black', zorder=1)
            else:
                ax[i].scatter(domains[i], values[i], label=name_with_params, color='black', zorder=1)

            # labeling
            ax[i].set_title(titles[i])
            ax[i].set_xlabel(xlabels[i])
            ax[i].set_ylabel(ylabels[i])
            ax[i].grid(grid)
            ax[i].legend(loc='upper center', bbox_to_anchor=(0.5, -0.05),
                         fancybox=True, shadow=True, ncol=1)
        
        plt.tight_layout()
        if show:
            plt.show()
        
###############################################################################
# Multivariate Base Class
###############################################################################
class GeneralMultivariate(Distribution):
    r"""Base class for multivariate and copula distributions."""
    @abstractmethod
    def _classify_params(self, params: dict, 
                         scalar_names: tuple = tuple(), 
                         vector_names: tuple = tuple(), 
                         shape_names: tuple = tuple(), 
                         symmetric_shape_names: tuple = tuple(), 
                         corr_like_shape_names: tuple = tuple(), 
                         non_symmetric_shape_names: tuple = tuple()) -> dict:
        r"""Classify the distribution parameters into scalars, vectors
        and shapes.
        
        Args:
            params (dict)
            scalar_names (tuple[str]): Tuple of scalar parameter names.
            vector_names (tuple[str]): Tuple of vector parameter names.
            symmetric_shape_names (tuple[str]): Tuple of symmetric shape 
                parameter names. Diagonal elements are included in the 
                parameter count.
            corr_like_shape_names (tuple[str]): Tuple of correlation 
                matrix like symmetric shape parameter names. Diagonal 
                elements are not included in the parameter count.
            non_symmetric_shape_names (tuple[str]): Tuple of 
                non-symmetric shape parameter names.
        """

        classifications = {
            "scalars": {name: params[name] for name in scalar_names},
            "vectors": {name: params[name] for name in vector_names},
            "shapes": {name: params[name] for name in shape_names},
            "symmetric_shapes": {name: params[name] for name in symmetric_shape_names},
            "corr_like_shapes": {name: params[name] for name in corr_like_shape_names},
            "non_symmetric_shapes": {name: params[name] for name in non_symmetric_shape_names},
        }
        return classifications

    @abstractmethod
    def _get_dim(self, params: dict) -> int:
        r"""Returns the number of dimensions of the distribution."""

    def _args_transform(self, params: dict) -> dict:
        classifications: dict = self._classify_params(params=params)  # todo: issue is here where scalars are being included. need to think about how to do this
        d: int = self._get_dim(params=params)

        # scalars
        transformed_scalars: dict = Distribution._scalar_transform(classifications["scalars"])
        
        # vectors
        transformed_vectors: dict = {k: jnp.asarray(v, dtype=float).reshape((d, 1)) 
                                     for k, v in classifications["vectors"].items()}
        
        # shapes
        transformed_shapes: dict = {k: jnp.asarray(v, dtype=float).reshape((d, d))
                                     for k, v in classifications["shapes"].items()}
        
        return {**transformed_scalars, **transformed_vectors, **transformed_shapes}
    
    def _get_num_params(self, params: dict) -> int:
        r"""Returns the number of parameters of the distribution.

        Returns:
            int: The number of parameters.
        """
        classifications: dict = self._classify_params(params=params)
        dim: int = self._get_dim(params=params)

        # scalars
        scalars: tuple = classifications["scalars"]
        n_scalars: int = len(scalars)
    
        # vectors
        vectors: dict = classifications["vectors"]
        n_vectors: int = len(vectors) * dim

        # symmetric shapes
        symmetric_shapes: dict = classifications["symmetric_shapes"]
        n_symm_shapes: int = len(symmetric_shapes) * (dim * (dim + 1) // 2)

        # correlation-like shapes
        corr_like_shapes: dict = classifications["corr_like_shapes"]
        n_corr_like_shapes: int = len(corr_like_shapes) * (dim * (dim - 1) // 2)

        # non-symmetric shapes
        non_symmetric_shapes: dict = classifications["non_symmetric_shapes"]
        n_non_symm_shapes: int = len(non_symmetric_shapes) * (dim ** 2)
        
        # total
        return n_scalars + n_vectors + n_symm_shapes + n_corr_like_shapes + n_non_symm_shapes
    

    @abstractmethod
    def support(self, params: dict) -> Array:
        r"""The support of the distribution is the subset of 
        multivariate x for which the pdf is non-zero. 

        Args:
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.
        
        Returns:
            Array: Array containing the support of each variable in
            the distribution, in an (d, 2) shape.
        """

    # sampling
    @abstractmethod
    def rvs(self, size: int, params: dict, 
            key: Array = DEFAULT_RANDOM_KEY) -> Array:
        r"""Generates random samples from the distribution.

        Note:
            If you intend to jit wrap this function, ensure that 'size' 
            is a static argument.

        Args:
            size (int): The size of the generated 
                output array of random numbers. Must be an integer. 
                Generates an (size, d) array of random numbers, where
                d is the number of dimensions inferred from the provided
                distribution parameters.
            params (dict): Parameters describing the distribution. See
                    the specific distribution class or the 'example_params' 
                    method for details.
            key (Array): The Key for random number generation.
        """
    
    @abstractmethod
    def sample(self, size: int, params: dict, 
               key: Array = DEFAULT_RANDOM_KEY) -> Array:
        r"""Generates random samples from the distribution.

        Note:
            If you intend to jit wrap this function, ensure that 'size' 
            is a static argument.

        Args:
            size (Scalar): The size / shape of the generated 
                output array of random numbers. Must be scalar. 
                Generates an (size, d) array of random numbers, where
                d is the number of dimensions inferred from the provided
                distribution parameters.
            key (Array): The Key for random number generation.
        """
        return super().sample(size=size, params=params, key=key)
    
    # metrics
    @abstractmethod
    def aic(self, x: ArrayLike, params: dict) -> float:
        k: int = self._get_num_params(params=params)
        return super().aic(k=k, x=x, params=params)

    @abstractmethod
    def bic(self, x: ArrayLike, params: dict) -> float:
        x, _, n, _ = _multivariate_input(x)
        k: int = self._get_num_params(params=params)
        return super().bic(k=k, n=n, x=x, params=params)

    # fitting
    @abstractmethod
    def fit(self, x: ArrayLike, *args, **kwargs) -> dict:
        r"""Fit the distribution to the input data.

        Note:
            Data must be in the shape (n, d) where n is the number of 
            samples and d is the number of dimensions.
        
        Args:
            x (ArrayLike): The input data to fit the distribution too.
                Must be in the shape (n, d) where n is the number of
                samples and d is the number of dimensions.
            kwargs: Additional keyword arguments to pass to the fit 
                method.
        
        Returns:
            dict: The fitted distribution parameters.
        """


class Multivariate(GeneralMultivariate):
    r"""Base class for multivariate distributions."""
    def _get_dim(self, params: dict) -> int:
        classifications: dict = self._classify_params(params)
        return jnp.asarray(list(classifications["vectors"].values())[0]).size

    def support(self, params: dict, 
                marginal_support: tuple = (-jnp.inf, jnp.inf), 
                *args, **kwargs) -> Array:
        d: int = self._get_dim(params=params)
        return jnp.concat([jnp.full((d, 1), marginal_support[0]), 
                           jnp.full((d, 1), marginal_support[1])], axis=1)
    
    @jit
    def _single_qi(self, carry: tuple, xi: jnp.ndarray) -> jnp.ndarray:
        mu, sigma_inv = carry
        return carry, lax.sub(xi, mu).T @ sigma_inv @ lax.sub(xi, mu)
    
    def _calc_Q(self, x: jnp.ndarray, mu: jnp.ndarray, sigma_inv: jnp.ndarray
                ) -> jnp.ndarray:
        r"""Calculates the Q vector (x - mu)^T @ sigma^-1 @ (x - mu)"""
        return lax.scan(f=self._single_qi, xs=x, 
                        init=(mu.flatten(), sigma_inv))[1]

    def logpdf(self, x: ArrayLike, params: dict) -> Array:
        r"""The log-probability density function (pdf) of the 
        distribution.

        Args:
            x (ArrayLike): The input at which to evaluate the log-pdf.
                Must be in the shape (n, d) where n is the number of
                samples and d is the number of dimensions.
            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.  

        Returns:
            Array: The log-pdf values.
        """
        return super().logpdf(x=x, params=params)


    def pdf(self, x: ArrayLike, params: dict) -> Array:
        r"""The probability density function (pdf) of the distribution.

        Args:
            x (ArrayLike): The input at which to evaluate the pdf.
                Must be in the shape (n, d) where n is the number of
                samples and d is the number of dimensions.

            params (dict): Parameters describing the distribution. See 
                the specific distribution class or the 'example_params' 
                method for details.  

        Returns:
            Array: The pdf values.
        """
        return super().pdf(x=x, params=params)
    
    @abstractmethod
    def _fit_copula(self, u: ArrayLike, corr_method: str = 'pearson', 
                    *args, **kwargs) -> dict:
        r"""Fits the copula distribution to the data.

        Note:
            If you intend to jit wrap this function, ensure that 
            'corr_method' is a static argument.
        
        Args:
            x (ArrayLike): The input data to fit the distribution too.
                Must be in the shape (n, d) where n is the number of
                samples and d is the number of dimensions.
            kwargs: Additional keyword arguments to pass to the fit 
                method.
        
        Returns:
            dict: The fitted copula distribution parameters.
        """
        u, _, n, d = _multivariate_input(u)
        mu: jnp.darray = jnp.zeros((d, 1))
        sigma: jnp.ndarray = corr(x=u, method=corr_method)
        return {'mu': mu, 'sigma': sigma, 'n': n, 'd': d, 'u': u}

    
class NormalMixture(Multivariate):
    r"""Base class for normal mixture distributions."""
    # sampling
    def _rvs(self, key, n: int, W: Array, mu: Array, gamma: Array, sigma: Array) -> Array:
        r"""Generates random samples from the normal-mixture 
        distribution."""
        d: int = mu.size

        m: jnp.ndarray = mu + W * gamma

        Z: jnp.ndarray = random.normal(key, shape=(d, n))
        A: jnp.ndarray = jnp.linalg.cholesky(sigma)
        r: jnp.ndarray = jnp.sqrt(W) * (A @ Z)
        return (m + r).T
    
    # stats
    def _stats(self, w_stats: dict, mu: Array, gamma: Array, sigma: Array) -> dict:
        mean: Array = mu + w_stats["mean"] * gamma
        cov: Array = w_stats["mean"] * sigma + w_stats["variance"] * jnp.outer(gamma, gamma)
        return {
            "mean": mean, 
            "cov": cov, 
            "skewness": gamma,}

    # fitting
    @abstractmethod
    def _ldmle_inputs(self, d: int) -> tuple:
        """Returns the input arguments for the low dimensional MLE.
        Specifically, the projection options containing the constraints
        and the initial guess."""
        pass
    
    def _general_fit(self, x: ArrayLike, d: int, loc: jnp.ndarray, shape: jnp.ndarray, 
                     reconstruct_func_id: int, lr: float, maxiter: int) -> dict:
        # optimisation constraints and initial guess
        projection_options, params0 = self._ldmle_inputs(d)

        # ADAM gradient descent
        res: dict = projected_gradient(f=self._ldmle_objective, x0=params0, 
                                       projection_method='projection_box', 
                                       projection_options=projection_options, 
                                       x=x, loc=loc, shape=shape, 
                                       reconstruct_func_id=reconstruct_func_id,
                                       lr=lr, maxiter=maxiter,)
        
        # reconstructing the parameters
        optimised_params_arr: jnp.ndarray = res['x']
        optimised_params: tuple = self._reconstruct_ldmle_func(
            func_id=reconstruct_func_id, params_arr=optimised_params_arr, 
            loc=loc, shape=shape)
        return optimised_params
    
    def _ldmle_objective(self, params_arr: jnp.ndarray, x: jnp.ndarray, 
                         loc: jnp.ndarray, shape: jnp.ndarray, reconstruct_func_id) -> Scalar:
        params: dict = self._reconstruct_ldmle_func(
            func_id=reconstruct_func_id, params_arr=params_arr, loc=loc, shape=shape)
        return -self._stable_logpdf(stability=1e-30, x=x, params=params).sum()
    
    @abstractmethod
    def _ldmle_inputs(self, d: int) -> tuple:
        """Returns the input arguments for the low dimensional MLE.
        Specifically, the projection options containing the constraints
        and the initial guess."""
        pass
    
    def fit(self, x: ArrayLike, cov_method: str = 'pearson', 
            lr: float = 1e-4, maxiter: int = 100) -> dict:
        r"""Fits the multivariate distribution to the data.

        Note:
            If you intend to jit wrap this function, ensure that 
            'cov_method' is a static argument.

        Algorithm:
            1. Estimate the sample mean vector and sample covariance 
                matrix, potentially using robust estimators.
            2. Remove the location and shape matrices from the 
                optimisation process, as these can be inferred from 
                scalar parameters and skewness.
            3. We maximise the log-likelihood using ADAM.

        Args:
            x: arraylike, data to fit the distribution to.
            cov_method: str, method to estimate the sample covariance 
                matrix, sigma. See copulax.multivariate.cov and/or 
                copulax.multivariate.corr for available methods.
            lr (float): Learning rate for optimization.
            maxiter (int): Maximum number of iterations for optimization.

        Returns:
            dict containing the fitted parameters.
        """
        # estimating the sample mean and covariance
        x, _, _, d = _multivariate_input(x)
        sample_mean: jnp.ndarray = jnp.mean(x, axis=0).reshape((d, 1))
        sample_cov: jnp.ndarray = cov(x=x, method=cov_method)

        # optimising
        return self._general_fit(
            x=x, d=d, loc=sample_mean, shape=sample_cov, 
            reconstruct_func_id=0, lr=lr, maxiter=maxiter,)
    
    @abstractmethod
    def _reconstruct_ldmle_params(self, params_arr: jnp.ndarray, 
                                  loc: jnp.ndarray, 
                                  shape: jnp.ndarray) -> dict:
        """Reconstructs the low dim MLE parameters from a flat array."""
        pass
    
    @abstractmethod
    def _reconstruct_ldmle_copula_params(self, params_arr: jnp.ndarray,
                                         loc: jnp.ndarray, 
                                         shape: jnp.ndarray) -> tuple:
        """Reconstructs the low dim MLE parameters from a flat array for 
        copula fitting."""
        pass

    def _reconstruct_ldmle_func(self, func_id: int, params_arr: jnp.ndarray, 
                                loc: jnp.ndarray, shape: jnp.ndarray) -> tuple:
        """Reconstructs the low dim MLE parameters from a flat array."""
        params_tuple: tuple = lax.cond(func_id == 0, 
                        self._reconstruct_ldmle_params, 
                        self._reconstruct_ldmle_copula_params, 
                        params_arr, loc, shape)
        return self._params_from_array(params_tuple)

    def _fit_copula(self, u: jnp.ndarray, corr_method: str, lr: float, 
                    maxiter: int):
        d: dict = super()._fit_copula(u, corr_method)
        return self._general_fit(
            x=d['u'], d=d['d'], loc=d['mu'], shape=d['sigma'], 
            reconstruct_func_id=1, lr=lr, maxiter=maxiter,)
        




"""File containing code for numerical approximations of univariate distribution 
functions."""
from jax._src.typing import ArrayLike, Array
from jax import jit, lax
import jax.numpy as jnp
from typing import Callable
from interpax import Interpolator1D

from copulax._src.optimize import projected_gradient, brent
from copulax._src.typing import Scalar


# def _ppf_optimizer(dist, q: ArrayLike, params: dict, x0: Scalar, 
#                    bounds: jnp.ndarray, lr: float, maxiter: int) -> Array:
    
#     def _ppf_func_single(xi: float, qi: float):
#         return jnp.abs((dist.cdf(x=xi, params=params) - qi)).reshape(())
    
#     min_val, max_val = bounds.reshape((2,1))
#     SCALE = 0.5
#     x0_small = jnp.max(jnp.array([x0 * (1-SCALE), min_val]))
#     x0_large = jnp.min(jnp.array([x0 * (1+SCALE), max_val]))
#     def _iter(carry, qi):
#         x0 = jnp.where(qi <= 0.5, x0_small, x0_large).reshape((1,))
#         res = projected_gradient(
#             f=_ppf_func_single, x0=x0, lr=lr, maxiter=maxiter, 
#             projection_method='projection_box', 
#             projection_options={'hyperparams': bounds}, qi=qi)
#         return carry, res['x']

#     _, x = jax.lax.scan(_iter, None, q)
#     return x.flatten()

# @jit
def _ppf_func_single(xi: float, qi: float, dist, params):
        return (dist.cdf(x=xi, params=params) - qi).reshape(())


def _ppf_optimizer(dist, q: ArrayLike, params: dict, bounds: tuple, maxiter: int) -> Array:
    # getting bound_maxiter
    factor: int = 10
    if q.dtype == jnp.int32 or q.dtype == int or q.dtype == jnp.uint32:
         width: int = 9
         resolution_power: int = 0
    elif q.dtype == jnp.int64 or q.dtype == jnp.uint or q.dtype == jnp.uint64:
         width: int = 19
         resolution_power: int = 0
    elif q.dtype == jnp.int4:
            width: int = 1
            resolution_power: int = 0
    elif q.dtype == jnp.int8 or q.dtype == jnp.uint8:
         width: int = 2
         resolution_power: int = 0
    elif q.dtype == jnp.int16 or q.dtype == jnp.uint16:
         width: int = 4
         resolution_power: int = 0
    elif q.dtype == jnp.float16:
         width: int = 4
         resolution_power: int = 3
    elif q.dtype == jnp.float32:
         width: int = 38
         resolution_power: int = 6
    else:
        # float64
        width: int = 308
        resolution_power: int = 15
    bound_maxiter: int = width + resolution_power

    def _left_iter(carry, _):
         left, right, qi, last_val = carry
         left, right = lax.cond(last_val > 0, lambda: (left * factor, left), lambda: (left, right))
         next_val = _ppf_func_single(xi=left, qi=qi, dist=dist, params=params)
         return (left, right, qi, next_val), _
    
    def _right_iter(carry, _):
         left, right, qi, last_val = carry
         left, right = lax.cond(last_val < 0, lambda: (right, right * factor), lambda: (left, right))
         next_val = _ppf_func_single(xi=right, qi=qi, dist=dist, params=params)
         return (left, right, qi, next_val), _
    
#     for qi in q:
#         # getting non-infinite left bound
#         left, right = bounds
#         non_inf_left = jnp.min(jnp.array([-factor, right]))
#         non_inf_left_val = _ppf_func_single(non_inf_left, qi, dist, params) 
#         left_res = lax.cond(jnp.isinf(left), lambda: lax.scan(_left_iter, (non_inf_left, right, qi, non_inf_left_val), length=bound_maxiter)[0], lambda: (left + 1e-5, right, qi, non_inf_left_val))
#         left, right = left_res[0], left_res[1]

#         # getting non-infinite right bound
#         non_inf_right = jnp.max(jnp.array([factor, left])) - 1e-5  # subtracting small value to avoid numerical issues
#         non_inf_right_val = _ppf_func_single(non_inf_right, qi, dist, params) 
#         left_res = lax.cond(jnp.isinf(right), lambda: lax.scan(_right_iter, (left, non_inf_right, qi, non_inf_right_val), length=bound_maxiter)[0], lambda: (left, right - 1e-5, qi, non_inf_right_val))
#         left, right = left_res[0], left_res[1]

#         # solving for root within bounds
#         new_qi = brent(method='bisection', g=_ppf_func_single, bounds=jnp.array([left, right]), maxiter=maxiter, qi=qi, dist=dist, params=params)
#         breakpoint()
        
    def _iter(carry, qi):
        eps = 1e-5

        # getting non-infinite left bound
        left, right = bounds
        non_inf_left = jnp.min(jnp.array([-factor, right]))
        non_inf_left_val = _ppf_func_single(non_inf_left, qi, dist, params) 
        left_res = lax.cond(jnp.isinf(left), lambda: lax.scan(_left_iter, (non_inf_left, right, qi, non_inf_left_val), length=bound_maxiter)[0], lambda: (left + eps, right, qi, non_inf_left_val))
        left, right = left_res[0], left_res[1]

        # getting non-infinite right bound
        non_inf_right = jnp.max(jnp.array([factor, left]))
        non_inf_right_val = _ppf_func_single(non_inf_right, qi, dist, params) 
        left_res = lax.cond(jnp.isinf(right), lambda: lax.scan(_right_iter, (left, non_inf_right, qi, non_inf_right_val), length=bound_maxiter)[0], lambda: (left, right - eps, qi, non_inf_right_val))
        left, right = left_res[0], left_res[1]

        # solving for root within bounds
        # using bisection method for robustness with gradients
        new_qi = brent(method='bisection', g=_ppf_func_single, bounds=jnp.array([left, right]), maxiter=maxiter, qi=qi, dist=dist, params=params)
        return carry, new_qi
    
    _, x = lax.scan(_iter, None, q)
    return x.flatten()


def _cubic_ppf(dist, q: ArrayLike, params: dict, 
                bounds: tuple, num_points: int,  maxiter: int) -> Array:
    # ensuring q has at least 3 elements
    if q.size < 3:
        # more efficient to use ppf_optimiser
        return dist.ppf(q=q, params=params, cubic=False, maxiter=maxiter)

    # clip q to (eps, 1-eps) to avoid numerical issues near bounds
    eps: float = 1e-5
    q_clipped: jnp.ndarray = jnp.clip(q, eps, 1-eps)

    # getting interpolation bounds
    q_range: jnp.ndarray = jnp.array([jnp.min(q_clipped), jnp.max(q_clipped)]
                                     ).reshape((2, 1))
    x_min, x_max = _ppf_optimizer(dist=dist, q=q_range, params=params, 
                                  bounds=bounds, maxiter=maxiter)

    # getting interpolation points
    x_range: jnp.ndarray = jnp.linspace(x_min, x_max, num_points).flatten()  
    cdf_values: jnp.ndarray = dist.cdf(x=x_range, params=params)

    # interpolating
    interpolator = Interpolator1D(x=cdf_values, f=x_range, method='monotonic', extrap=True)
    interpolated_clipped_values: jnp.ndarray = interpolator(q_clipped)
    interpolated_values: jnp.ndarray = jnp.where(q <= eps, bounds[0], interpolated_clipped_values)
    interpolated_values = jnp.where(q >= 1-eps, bounds[1], interpolated_values)
    return interpolated_values
    


def _ppf(dist, q: ArrayLike, params: dict, cubic: bool, 
         num_points: int, maxiter: int) -> Array:
    q = q.flatten()
    bounds: tuple = dist._support(params)
    if cubic:
        x: jnp.ndarray = _cubic_ppf(dist=dist, q=q, params=params, 
                                     bounds=bounds, num_points=num_points, 
                                     maxiter=maxiter)
    else:
        x: jnp.ndarray = _ppf_optimizer(dist=dist, q=q, params=params, 
                                        bounds=bounds, maxiter=maxiter)
        
    x = jnp.where(jnp.logical_and(x >= bounds[0], x <= bounds[1]), x, jnp.nan)
    x = jnp.where(q == 0, bounds[0], x)
    return jnp.where(q == 1, bounds[1], x)



"""File containing the copulAX implementation of the multivariate 
generalized hyperbolic (GH) distribution."""
import jax.numpy as jnp
from jax import lax, random, jit
from jax._src.typing import ArrayLike, Array
from jax.scipy import special

from copulax._src._distributions import NormalMixture
from copulax._src.typing import Scalar
from copulax._src.multivariate._utils import _multivariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.multivariate._shape import cov, _corr
from copulax._src.univariate.gig import gig
from copulax.special import kv


class MvtGH(NormalMixture):
    r"""The multivariate generalized hyperbolic (GH) distribution is a
    generalization of the univariate GH distribution to d > 1 
    dimensions. This is a flexible, continuous 6-parameter family of 
    distributions that can model a variety of data behaviors, including 
    heavy tails and skewness. It contains a number of popular 
    distributions as special cases, including the multivariate normal, 
    multivariate student-t and multivariate skewed-T distributions.

    We adopt the parameterization used by McNeil et al. (2005)
    """
    def _classify_params(self, params: dict) -> tuple:
        # return (lamb, chi, psi,), (mu, gamma), (sigma,)
        return super()._classify_params(
            params=params, scalar_names=('lambda', 'chi', 'psi'),
            vector_names=('mu', 'gamma'), shape_names=('sigma',),
            symmetric_shape_names=('sigma',))
    
    def _params_dict(self, lamb: Scalar, chi: Scalar, psi: Scalar,
                     mu: ArrayLike, gamma: ArrayLike, sigma: ArrayLike) -> dict:
        d: dict = {"lambda": lamb, "chi": chi, "psi": psi, "mu": mu, 
                   "gamma": gamma, "sigma": sigma}
        return self._args_transform(d)
    
    def _params_to_tuple(self, params: dict) -> tuple:
        params = self._args_transform(params)
        return (params["lambda"], params["chi"], params["psi"],
                params["mu"], params["gamma"], params["sigma"])
    
    def example_params(self, dim: int = 3, *args, **kwargs) -> dict:
        r"""Example parameters for the multivariate GH distribution.
        
        This is a six parameter family, defined by the scalar parameters 
        `lambda`, `chi`, `psi`, the location vector `mu`, the 
        skewness vector `gamma` and the shape matrix `sigma`.

        Args:
            dim: int, number of dimensions of the multivariate GH 
                distribution. Default is 3.
        """
        return self._params_dict(lamb=0.0, chi=1.0, psi=1.0, 
                                 mu=jnp.zeros((dim, 1)), 
                                 gamma=jnp.zeros((dim, 1)), 
                                 sigma=jnp.eye(dim, dim))
    
    def support(self, params: dict) -> Array:
        return super().support(params=params)
    
    # @jit
    # def _single_hi(self, carry: tuple, xi: jnp.ndarray) -> jnp.ndarray:
    #     mu, gamma, sigma_inv = carry
    #     return carry, lax.sub(xi, mu).T @ sigma_inv @ gamma
    
    # def _calc_H(self, x: ArrayLike, mu: ArrayLike, gamma: ArrayLike, sigma_inv: ArrayLike) -> Array:
    #     r""""Calculates the H vector (x - mu).T @ sigma^-1 @ gamma."""
    #     return lax.scan(f=self._single_hi, xs=x, 
    #                     init=(mu.flatten(), gamma.flatten(), sigma_inv))[1]
    
    def _stable_logpdf(self, stability: Scalar, x: ArrayLike, 
                       params: dict) -> Array:
        x, yshape, n, d = _multivariate_input(x)
        lamb, chi, psi, mu, gamma, sigma = self._params_to_tuple(params)
        # sigma: Array = _corr._rm(sigma, 1e-5)

        sigma_inv: Array = jnp.linalg.inv(sigma)
        Q: Array = chi + self._calc_Q(x=x, mu=mu, sigma_inv=sigma_inv)
        R: Array = psi + gamma.T @ sigma_inv @ gamma
        QR: Array = Q * R
        # H: Array = self._calc_H(x=x, mu=mu, gamma=gamma, sigma_inv=sigma_inv)
        H: Array = (x @ sigma_inv @ gamma).flatten()
        log_det_sigma: Scalar = jnp.linalg.slogdet(sigma)[1]
        s: Scalar = lamb - d / 2

        log_c: Scalar = (0.5 * lamb * lax.log((chi / (psi + stability)) + stability) - s * lax.log(R + stability) - 0.5 * d * lax.log(2 * jnp.pi) - 0.5 * log_det_sigma)
        
        logpdf: Array = (log_c 
                         + lax.log(kv(s, lax.sqrt(QR)) + stability) 
                         + H 
                         - 0.5 * s * (lax.log(QR + stability)))
        return logpdf.reshape(yshape)
    
    # sampling
    def rvs(self, size: int, params: dict, 
            key: ArrayLike=DEFAULT_RANDOM_KEY) -> Array:
        lamb, chi, psi, mu, gamma, sigma = self._params_to_tuple(params)

        key, subkey = random.split(key)
        W: Array = gig.rvs(size=(size,), key=key, 
                           params={'lambda': lamb, 'chi': chi, 'psi': psi})
        return super()._rvs(key=subkey, n=size, W=W, mu=mu, gamma=gamma, sigma=sigma)
    
    # stats
    def stats(self, params: dict) -> dict:
        lamb, chi, psi, mu, gamma, sigma = self._params_to_tuple(params)
        gig_stats = gig.stats(params={'lambda': lamb, 'chi': chi, 'psi': psi})
        return self._stats(w_stats=gig_stats, mu=mu, gamma=gamma, sigma=sigma)
    
    # fitting
    def _ldmle_inputs(self, d):
        lc = jnp.full((d + 3,1), -jnp.inf)
        uc = jnp.full((d + 3,1), jnp.inf)
        
        key1, key2 = random.split(DEFAULT_RANDOM_KEY)
        key2, key3 = random.split(key2)
        params0 = jnp.array([random.normal(key1), *lax.exp(random.normal(key2, (2,))), *random.normal(key3, (d,))]).flatten()
        return {'hyperparams': (lc, uc)}, params0
    
    def _reconstruct_ldmle_params(self, params_arr, loc, shape):
        d: int = loc.size
        scalars = lax.dynamic_slice_in_dim(params_arr, 0, 3)
        lamb, chi_, psi_ = scalars
        chi, psi = jnp.exp(chi_), jnp.exp(psi_)
        gamma: Array = lax.dynamic_slice_in_dim(params_arr, 3, d).reshape((d, 1))
        gig_stats: dict = gig.stats(params={'lambda': lamb, 'chi': chi, 'psi': psi})

        mu: Array = loc - gig_stats["mean"] * gamma
        sigma_: Array = (shape - gig_stats["variance"] * jnp.outer(gamma, gamma)) / gig_stats["mean"]
        sigma: Array = _corr._rm_incomplete(sigma_, 1e-5)
        return lamb, chi, psi, mu, gamma, sigma
    

    def _reconstruct_ldmle_copula_params(self, params_arr, loc, shape):
        d: int = loc.size
        scalars = lax.dynamic_slice_in_dim(params_arr, 0, 3)
        lamb, chi_, psi_ = scalars
        chi, psi = jnp.exp(chi_), jnp.exp(psi_)
        gamma = lax.dynamic_slice_in_dim(params_arr, 3, d).reshape((d, 1))
        return lamb, chi, psi, loc, gamma, shape


mvt_gh = MvtGH("Mvt-GH")



"""File containing code for numerical approximations of univariate distribution 
functions."""
from jax._src.typing import ArrayLike, Array
import jax.numpy as jnp
from typing import Callable

from copulax._src.univariate._utils import _univariate_input


def _cdf_approx(key: ArrayLike, rvs_func: Callable, x: ArrayLike, params: tuple,
                num_points: int = 500) -> Array:
    """Approximate the cdf of a distribution by evaluating it at a set of 
    points. Uses a linear interpolation to reduce the required number of 
    similated random variables.

    Args:
        key: PRNGKey for random number generation.
        rvs_func: Function which simulates univariate random variables.
        x: arraylike, value(s) at which to evaluate the cdf. This must be a
        one-dimensional array.
        params: Parameters of the distribution.
        num_points: Number of simulated random points at which to evaluate the 
        cdf.

    Returns:
        Array of cdf values.
    """
    x: jnp.ndarray = _univariate_input(x)

    # performing linear interpolation 
    xp: jnp.ndarray = rvs_func(*params, key=key, shape=(num_points,)) 
    xp = xp.sort(axis=0)
    fp = jnp.where(xp <= xp.reshape((num_points, 1)), 1.0, 0.0).mean(axis=1)
    return jnp.interp(x=x, xp=xp, fp=fp, left=0.0, right=1.0)


def _ppf_approx(key: ArrayLike, rvs_func: Callable, q: ArrayLike, params: tuple,
                num_points: int = 500) -> Array:
    """Approximate the percent point function (inverse of cdf) of a 
    distribution by evaluating it at a set of points. Uses a linear 
    interpolation to reduce the required number of simulated random variables.
    
    Args:
        key: PRNGKey for random number generation.
        rvs_func: Function which simulates univariate random variables.
        q: arraylike, value(s) at which to evaluate the ppf.
        params: Parameters of the distribution.
        num_points: Number of simulated random points at which to evaluate the 
        cdf.
    """
    q: jnp.ndarray = _univariate_input(q)

    # performing linear interpolation 
    fp: jnp.ndarray = rvs_func(key, (num_points,), *params) 
    fp = fp.sort(axis=0)
    xp = jnp.where(fp <= fp.reshape((num_points, 1)), 1.0, 0.0).mean(axis=1)
    return jnp.interp(x=q, xp=xp, fp=fp)



"""Implements a jit-able, jax-differentiable version of numerical univariate cdf integration."""
from jax import numpy as jnp
from jax import grad, lax
from typing import Callable
from quadax import quadgk, quadcc


from copulax._src.univariate._utils import _univariate_input


METHOD: Callable = quadgk


def _cdf_single_x(pdf_func: Callable, lower_bound: float, xi: float, params_array) -> float:
    cdf_vals, info = METHOD(fun=pdf_func, interval=[lower_bound, xi], args=params_array, )
    return cdf_vals.reshape(())


def _cdf(dist, x: jnp.ndarray, params: dict) -> jnp.ndarray:
    # adding right bound to the cdf integral
    x, xshape = _univariate_input(x)
    xsize = x.size
    lower_bound, upper_bound = dist.support(params)
    x = jnp.append(x, upper_bound.reshape((1, 1)), axis=0)

    params_array: jnp.ndarray = dist._params_to_array(params)

    def _iter(carry, xi):
        cdf_i = _cdf_single_x(dist._pdf_for_cdf, lower_bound, xi, params_array)
        return carry, cdf_i
    
    _, cdf_raw_ = lax.scan(_iter, None, x.flatten())

    # ensuring the cdf is scaled to be between 0 and 1
    cdf_raw = lax.dynamic_slice_in_dim(cdf_raw_, 0, xsize, axis=0)
    scale = lax.dynamic_slice_in_dim(cdf_raw_, xsize, 1, axis=0)
    cdf_adj = cdf_raw / scale
    cdf_adj = jnp.where(cdf_adj> 1.0, 1.0, cdf_adj)
    cdf_adj = jnp.where(cdf_adj < 0.0, 0.0, cdf_adj)

    return cdf_adj.reshape(xshape)


def _cdf_fwd(dist, cdf_func: Callable, x: jnp.ndarray, params: dict):
    x, xshape = _univariate_input(x)

    def cdf_single(xi, params):
        return cdf_func(xi, params).reshape(())

    def iter(carry, xi):
        params_grad_i = grad(cdf_single, argnums=1)(xi, params)
        return carry, params_grad_i

    _, param_grads = lax.scan(iter, None, x.flatten())
    pdf_values = dist.pdf(x=x, params=params).reshape(xshape)
    return cdf_func(x=x, params=params).reshape(xshape), (pdf_values, param_grads)


def cdf_bwd(res, g):
    xshape = res[0].shape
    g = g.reshape(xshape)
    x_grad = res[0] * g
    param_grads: dict = {key: jnp.mean(jnp.nan_to_num(val, 0.0) * g) 
                         for key, val in res[1].items()}  # average parameter gradients over x
    return x_grad, param_grads



from jax import lax
import jax.numpy as jnp
from jax._src.typing import ArrayLike, Array
from quadax import quadgk
from jax.scipy import special
from typing import Callable
import jax

# from copulax.univariate import student_t, normal, lognormal
# from copulax._src.univariate._utils import DEFAULT_RANDOM_KEY

# sample_dist, SCALE = normal, 2.0


# sample_dist, SCALE = student_t, 2.0
# sample_dist, SCALE = lognormal, 1.0


# def _kv_integrand(t: Array, v: float, x: Array) -> Array:
#     cosh_t = lax.cosh(t)
#     cosh_tv = lax.cosh(lax.mul(v, t))
#     inner = -x * cosh_t
#     exp = lax.exp(inner)

#     integrand = exp * cosh_tv

#     integrand = jnp.where(jnp.isinf(cosh_t), 0.0, integrand)
#     integrand = jnp.where(jnp.isinf(cosh_tv), 0.0, integrand)
#     integrand = jnp.where(jnp.isinf(exp), 0.0, integrand)
#     # integrand = jnp.where(jnp.isinf(inner), 0.0, integrand)
#     return integrand


# def _kv_integrand(t: Array, v: float, x: Array) -> Array:
#     cosh_t = lax.cosh(t)
#     cosh_t = jnp.where(jnp.isinf(cosh_t), 0.0, cosh_t)

#     cosh_tv = lax.cosh(lax.mul(v, t))
#     cosh_tv = jnp.where(jnp.isinf(cosh_tv), 0.0, cosh_tv)

#     inner = -x * cosh_t
#     exp = lax.exp(inner)
#     exp = jnp.where(jnp.isinf(exp), 0.0, exp)

#     integrand = exp * cosh_tv

#     # integrand = jnp.where(jnp.isinf(cosh_t), 0.0, integrand)
#     # integrand = jnp.where(jnp.isinf(cosh_tv), 0.0, integrand)
#     # integrand = jnp.where(jnp.isinf(exp), 0.0, integrand)
#     # integrand = jnp.where(jnp.isinf(inner), 0.0, integrand)
#     return integrand


def _kv_integrand(w: Array, v: float, x: Array) -> Array:
    STABILITY_TERM = 0 #1e-15
    frac = jnp.pow(w + STABILITY_TERM, -1)
    inner = -0.5*x * (w + frac)
    exp = lax.exp(inner)
    return 0.5 * lax.pow(w + STABILITY_TERM, v - 1.0) * exp


# def _kv_integrand(w: Array, v: float, x: Array) -> Array:
#     STABILITY_TERM = 0#1e-15
#     bottom = lax.pow(lax.pow(w, 2) + lax.pow(x, 2), v + 0.5)
#     vals = lax.cos(w) / bottom
#     vals = jnp.where(jnp.isinf(w), 0.0, vals)
#     vals = jnp.where(jnp.isinf(bottom), 0.0, vals)
#     return vals

# def _kv_integrand(w: Array, v: float, x: Array) -> Array:


def _kv_single_x(v: float, xi: float) -> float:
    # from jax import config
    # config.update("jax_debug_nans", True)
    # config.update("jax_disable_jit", True)

    kv_val, info = quadgk(_kv_integrand, interval=[0.0, jnp.inf], args=(v, xi))
    return kv_val.reshape(())


def kv(v: float, x: ArrayLike) -> Array:
    r"""Modified Bessel function of the second/third kind.

    Args:
        v: float, order of the Bessel function.
        x: arraylike, value(s) at which to evaluate the function.

    Returns:
        array of values.
    """
    v = jnp.asarray(jnp.abs(v), dtype=float)
    x = jnp.asarray(x, dtype=float)
    xshape = x.shape
    x = x.flatten()

    def _iter(carry, xi):
        kv_i = _kv_single_x(v, xi)
        return carry, kv_i
    
    _, kv_raw = lax.scan(_iter, None, x)

    kv_adj = jnp.where(x < 0, jnp.nan, kv_raw)
    return kv_adj.reshape(xshape)


def kv_x_to_0(v: float, x: ArrayLike) -> Array:
    r"""Modified Bessel function of the second/third kind.

    Args:
        v: float, order of the Bessel function.
        x: arraylike, value(s) at which to evaluate the function.

    Returns:
        array of values.
    """
    return special.gamma(v) * jnp.power(2, v - 1) * jnp.power(x, -v)


def kv_x_to_inf(v: float, x: ArrayLike) -> Array:
    return 0.0



# def kv_asymptotic_(v: float, x: ArrayLike) -> Array:
#     r"""Modified Bessel function of the second/third kind.

#     Incorporates limit approximations for x -> 0 and x -> inf. 
#     This allows for a more stable evaluation of gradients at the expense of 
#     precision loss. If the value of the modified Bessel function of the second 
#     kind is important, use kv. If gradient is important use kv_asymptotic.

#     Args:
#         v: float, order of the Bessel function.
#         x: arraylike, value(s) at which to evaluate the function.

#     Returns:
#         array of values.
#     """
#     v = jnp.asarray(jnp.abs(v), dtype=float)
#     x = jnp.asarray(x, dtype=float)
    
#     # preventing nans resulting from undesired gradient propagation via conditioning
#     raw_val = jax.lax.stop_gradient(kv(v, x))

#     # hacky way of implementing if, elif, else statements
#     index = 0
#     index = lax.cond(raw_val > 10.0, lambda _: 1, lambda _: index, None)
#     index = lax.cond(raw_val < 1e-10, lambda _: 2, lambda _: index, None)

#     return lax.switch(index, (kv, kv_x_to_0, kv_x_to_inf), v, x)


@jax.custom_gradient
def kv_asymptotic_single(v: float, xi: float) -> float:
    v = jnp.asarray(jnp.abs(v), dtype=float)
    xi = jnp.asarray(xi, dtype=float)
    
    # preventing nans resulting from undesired gradient propagation via conditioning
    kv_val = jax.lax.stop_gradient(kv(v, xi))

    # hacky way of implementing if, elif, else statements
    index = 0
    index = lax.cond(kv_val > 10.0, lambda _: 1, lambda _: index, None)
    index = lax.cond(kv_val < 1e-10, lambda _: 2, lambda _: index, None)

    # calculating the gradient
    kv_grad_func: Callable = (lambda v_, x_:  lax.switch(index, (_kv_single_x, kv_x_to_0, kv_x_to_inf), v_, x_))
    kv_grad: jnp.ndarray = jax.grad(kv_grad_func, argnums=(0, 1))(v, xi)
    return kv_val, lambda g: (g * kv_grad[0], g * kv_grad[1])


def kv_asymptotic(v: float, x: ArrayLike) -> Array:
    r"""Modified Bessel function of the second/third kind.

    Allows for a more stable evaluation of gradients at the expense of 
    precision loss by using asymptotic forms of the Bessel function when 
    calculating gradients. The kv value itself is not effected.

    Args:
        v: float, order of the Bessel function.
        x: arraylike, value(s) at which to evaluate the function.

    Returns:
        array of values.
    """
    v = jnp.asarray(jnp.abs(v), dtype=float)
    x = jnp.asarray(x, dtype=float)
    xshape = x.shape
    x = x.flatten()

    def _iter(carry, xi):
        kv_i = kv_asymptotic_single(v, xi)
        return carry, kv_i
    
    _, kv_vals = lax.scan(_iter, None, x)
    return kv_vals.reshape(xshape)

# maybe use personalised grad func 

    # return lax.cond(x <= 1e-5, kv_x_to_0, kv_, v, x)

    # return kv_(v, x)


# def _kv_integrand(v: float, x: Array, t: Array) -> Array:
#     cosh_t = lax.cosh(t)
#     cosh_tv = lax.cosh(lax.mul(v, t))
#     inner = -x * cosh_t
#     exp = lax.exp(inner)

#     integrand = exp * cosh_tv

#     integrand = jnp.where(jnp.isinf(cosh_t), 0.0, integrand)
#     integrand = jnp.where(jnp.isinf(cosh_tv), 0.0, integrand)
#     integrand = jnp.where(jnp.isinf(exp), 0.0, integrand)
#     # integrand = jnp.where(jnp.isinf(inner), 0.0, integrand)
#     return integrand


# def kv(v: float, x: ArrayLike, key=DEFAULT_RANDOM_KEY, num_points: int = 100) -> Array:
#     r"""Modified Bessel function of the second/third kind.

#     Approximates the indefinate integral using monte-carlo integration.

#     Args:
#         v: float, order of the Bessel function.
#         x: arraylike, value(s) at which to evaluate the function.
#         key: random.PRNGKey, random key for generating samples to use in 
#         monte-carlo integration.
#         num_points: int, number of random points to sample in monte-carlo 
#         integration.

#     Returns:
#         array of values.
#     """
#     eps = 1e-6
#     v = jnp.asarray(jnp.abs(v), dtype=float)
#     x = jnp.asarray(x, dtype=float)
#     xshape = x.shape
#     x = x.reshape((x.size, 1))
#     num_points: int = int(num_points)

#     t: Array = jnp.abs(sample_dist.rvs(key=key, shape=(num_points,)))
#     t = jnp.where(t < eps, eps, t)
#     samples = _kv_integrand(v, x, t)

#     pdf_vals = sample_dist.pdf(t).flatten() * SCALE
#     pdf_vals = jnp.where(pdf_vals < eps, eps, pdf_vals)
#     values = samples / pdf_vals
#     values = values.mean(axis=1).reshape(xshape)

#     values = jnp.where(x < 0, jnp.nan, values)
#     return values.reshape(xshape)





# def _kv_integrand(v: float, x: Array, t: Array) -> Array:
#     # top = lax.cos(t)
#     top = jnp.cos(t)
#     # top = 1.0
#     bottom = lax.pow(lax.pow(t, 2) + lax.pow(x, 2), v + 0.5)
#     return top / bottom


# def kv(v: float, x: ArrayLike, key=random.PRNGKey(0), num_points: int = 250) -> Array:
#     r"""Modified Bessel function of the second/third kind.

#     Approximates the indefinate integral using monte-carlo integration.

#     Args:
#         v: float, order of the Bessel function.
#         x: arraylike, value(s) at which to evaluate the function.
#         key: random.PRNGKey, random key for generating samples to use in 
#         monte-carlo integration.
#         num_points: int, number of random points to sample in monte-carlo 
#         integration.

#     Returns:
#         array of values.
#     """
#     eps = 1e-6
#     v = jnp.asarray(jnp.abs(v), dtype=float)
#     x = jnp.asarray(x, dtype=float)
#     xshape = x.shape
#     x = x.reshape((x.size, 1))
#     num_points: int = int(num_points)

#     t: Array = jnp.abs(sample_dist.rvs(key=key, shape=(num_points,)))
#     t = jnp.where(t < eps, eps, t)
#     c = lax.exp(lax.lgamma(v + 0.5)) * lax.pow(2.0 * x, v) / lax.sqrt(jnp.pi)
#     samples = lax.mul(c, _kv_integrand(v, x, t))

#     pdf_vals = sample_dist.pdf(t).flatten() * scale
#     pdf_vals = jnp.where(pdf_vals < eps, eps, pdf_vals)
#     values = samples / pdf_vals
#     values = values.mean(axis=1)[:, None]

#     values = jnp.where(x < 0, jnp.nan, values)
#     # if jnp.isnan(values).any():
#     #     breakpoint()
#     return values.reshape(xshape)


"""File containing the copulAX implementation of the lognormal distribution."""
import jax.numpy as jnp
from jax._src.typing import ArrayLike, Array

from copulax._src._distributions import Univariate
from copulax._src.typing import Scalar
from copulax._src.univariate._utils import _univariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.univariate.normal import normal


class LogNormal(Univariate):
    r"""The log-normal distribution of X is one where Y = log(X) is normally 
    distributed. It is a continuous 2 parameter family of distributions.
    
    https://en.wikipedia.org/wiki/Log-normal_distribution
    """
    def _params_to_tuple(self, params: dict):
        return normal._params_to_tuple(params)

    def example_params(self, *args, **kwargs):
        r"""Example parameters for the log-normal distribution.
        
        This is a two parameter family, with the log-normal being 
        defined by the mean and standard deviation of its transformed 
        distribution Y = log(X).
        """
        return normal.example_params()
    
    @classmethod
    def _support(cls, *args, **kwargs) -> tuple:
        return 0.0, jnp.inf
    
    def logpdf(self, x: ArrayLike, params: dict) -> Array:
        x, xshape = _univariate_input(x)
        x = x.reshape(xshape)
        return normal.logpdf(x=jnp.log(x), params=params) - jnp.log(x)
    
    def logcdf(self, x: ArrayLike, params: dict) -> Array:
        return normal.logcdf(x=jnp.log(x), params=params)
    
    def cdf(self, x: ArrayLike, params: dict) -> Array:
        return normal.cdf(x=jnp.log(x), params=params)

    # ppf    
    # def _get_x0(self, params: dict):
    #     return normal._get_x0(params=params)
    
    def _ppf(self, q: ArrayLike, params: dict, *args, **kwargs) -> Array:
        return jnp.exp(normal._ppf(q=q, params=params, *args, **kwargs))
    
    # sampling
    def rvs(self, size: tuple | Scalar, params: dict, key: Array = DEFAULT_RANDOM_KEY) -> Array:
        return jnp.exp(normal.rvs(size=size, key=key, params=params))
    
    # stats
    def stats(self, params: dict) -> dict:
        mu, sigma = self._params_to_tuple(params)

        mean: float = jnp.exp(mu + jnp.pow(sigma, 2) / 2)
        median: float = jnp.exp(mu)
        mode: float = jnp.exp(mu - jnp.pow(sigma, 2))
        variance: float = (jnp.exp(jnp.pow(sigma, 2)) - 1) * jnp.exp(2 * mu + jnp.pow(sigma, 2))
        std: float = jnp.sqrt(variance)
        skewness: float = (jnp.exp(jnp.pow(sigma, 2)) + 2) * jnp.sqrt(jnp.exp(jnp.pow(sigma, 2)) - 1)
        kurtosis: float = jnp.exp(4 * jnp.pow(sigma, 2)) + 2 * jnp.exp(3 * jnp.pow(sigma, 2)) + 3 * jnp.exp(2 * jnp.pow(sigma, 2)) - 6

        return self._scalar_transform({
            'mean': mean, 
            'median': median, 
            'mode': mode, 
            'variance': variance, 
            'std': std,
            'skewness': skewness, 
            'kurtosis': kurtosis})
    
    # fitting
    def fit(self, x: ArrayLike, *args, **kwargs) -> dict:
        return normal.fit(jnp.log(x))
    

lognormal = LogNormal("LogNormal")


import pytest
import numpy as np

from copulax.univariate import *


NUM_SAMPLES: int = 100
np.random.seed(0)


@pytest.fixture(scope='package', autouse=True)
def continuous_data():
    # creating some skewed, strictly positive data
    normal = np.random.normal(0, 1, NUM_SAMPLES)
    uniform = np.random.uniform(0, 1, NUM_SAMPLES)
    sample = normal * uniform
    sample[sample > 0] = sample[sample > 0] * 3 # adding skew
    sample = sample + 3 # shift the data to the right
    return sample[sample > 0]


@pytest.fixture(scope='package', autouse=True)
def discrete_data():
    # creating some discrete data
    sample = np.random.randint(0, 10, NUM_SAMPLES)
    return sample


@pytest.fixture(scope='package', autouse=True)
def continuous_dists():
    return {gamma, gh, gig, ig, lognormal, normal, skewed_t, student_t, uniform}


@pytest.fixture(scope='package', autouse=True)
def non_inverse_transform_dists():
    return {gamma, gh, gig, ig, lognormal, normal, skewed_t, student_t, uniform}


@pytest.fixture(scope='package', autouse=True)
def inverse_transform_dists():
    return {}


@pytest.fixture(scope='package', autouse=True)
def uniform_data():
    eps: float = 1e-6
    return np.random.uniform(eps, 1-eps, 10)


from typing import Callable
from jax._src.typing import Array
import jax.numpy as jnp
from jax import lax, random


def inverse_transform_sampling(ppf_func: Callable, shape: tuple, params: dict, key: Array) -> Array:
    """Generate random samples using the inverse transform sampling method.

    Args:
        ppf_func: The percent point function of the distribution.
        shape: The shape of the output array.
        params: The parameters of the distribution.

    Returns:
        Array: The generated random samples.
    """
    # num_samples = jnp.asarray(shape).prod()
    eps: float = 1e-2
    u: jnp.ndarray = random.uniform(key=key, shape=shape, minval=eps, maxval=1-eps)
    return ppf_func(q=u, **params).reshape(shape)


def mean_variance_sampling(key: Array, W: jnp.ndarray, shape: tuple, mu: float, sigma: float, gamma: float) -> jnp.ndarray:
    Z: jnp.ndarray = random.normal(key=key, shape=shape)
    m: jnp.ndarray = mu + W * gamma
    s: jnp.ndarray = lax.sqrt(W) * sigma * Z
    s = lax.mul(lax.sqrt(W) * sigma, Z)
    X: jnp.ndarray = m + s
    return X.reshape(shape)


import numpy as np
import jax
import jax.numpy as jnp

from copulax.special import kv



def test_kv():
    # checking values and gradients
    kv_vg = jax.jit(jax.value_and_grad(kv, argnums=(0, 1)))
    for v in np.linspace(-10, 10, 100):
        for x in np.linspace(1e-10, 10, 100):
            val, grad = kv_vg(v, x)

            assert val >= 0, f"kv not positive for v={v}, x={x}"
            assert not np.isnan(val), f"kv contains NaNs for v={v}, x={x}"
            assert (not np.isnan(grad[0])) and (not np.isnan(grad[1])), f"kv gradient contains NaNs for v={v}, x={x}"

    # checking shape
    xs = [jnp.array([[1.0, 2.0, 3.0], [2.3, 3.4, 5.5]]), jnp.array([1.0, 2.0, 3.0]), jnp.array(1.0)]
    for x in xs:
        val = kv(0.5, x)
        assert val.shape == x.shape, f"kv shape mismatch for x={x}"


"""File containing the copulAX implementation of the multivariate 
student-t distribution."""
import jax.numpy as jnp
from jax import lax, random, jit
from jax._src.typing import ArrayLike, Array
from jax.scipy import special

from copulax._src._distributions import NormalMixture
from copulax._src.typing import Scalar
from copulax._src.multivariate._utils import _multivariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.multivariate._shape import cov
from copulax._src.univariate.ig import ig


class MvtStudentT(NormalMixture):
    r"""The multivariate student-t distribution is a generalization of 
    the univariate student-t distribution to d > 1 dimensions.
    
    https://en.wikipedia.org/wiki/Multivariate_t-distribution

    :math:`\mu` is the mean vector and :math:`\sigma` the shape matrix,
    which for this parameterization is not the variance-covariance 
    matrix of the data distribution. :math:`\nu` is the degrees of 
    freedom parameter.
    """
    def _classify_params(self, params: dict) -> dict:
        return super()._classify_params(
            params=params, scalar_names=('nu',), vector_names=('mu',), 
            shape_names=('sigma',), symmetric_shape_names=('sigma',))
    
    def _params_dict(self, nu: Scalar, mu: ArrayLike, sigma: ArrayLike) -> dict:
        d: dict = {"nu": nu, "mu": mu, "sigma": sigma}
        return self._args_transform(d)
    
    def _params_to_tuple(self, params: dict) -> tuple:
        params = self._args_transform(params)
        return params["nu"], params["mu"], params["sigma"]
    
    def example_params(self, dim: int = 3, *args, **kwargs) -> dict:
        r"""Example parameters for the multivariate student-t distribution.

        This is a three parameter family, defined by the degrees of 
        freedom scalar `nu`, the mean / location vector `mu` and the
        shape matrix `sigma`.

        Args:
            dim: int, number of dimensions of the multivariate student-t
                distribution. Default is 3.
        """
        return self._params_dict(nu=2.5, mu=jnp.zeros((dim, 1)), 
                                 sigma=jnp.eye(dim, dim))
    
    def support(self, params: dict) -> Array:
        return super().support(params=params)
    
    def _stable_logpdf(self, stability: Scalar, x: ArrayLike, 
                       params: dict) -> Array:
        x, yshape, n, d = _multivariate_input(x)
        nu, mu, sigma = self._params_to_tuple(params)

        s: Scalar = 0.5 * (nu + d)
        sigma_inv: Array = jnp.linalg.inv(sigma)
        Q: Array = self._calc_Q(x=x, mu=mu, sigma_inv=sigma_inv)
        log_det_sigma: Scalar = jnp.linalg.slogdet(sigma)[1]
        logpdf: Array = (lax.lgamma(s) 
                         - lax.lgamma(0.5 * nu) 
                         - 0.5 * d * jnp.log(jnp.pi * nu + stability) 
                         - 0.5 * log_det_sigma 
                         - s * jnp.log1p(Q / nu))
        return logpdf.reshape(yshape)
    
    # sampling
    def rvs(self, size: int, params: dict, 
            key: ArrayLike=DEFAULT_RANDOM_KEY) -> Array:
        nu, mu, sigma = self._params_to_tuple(params)

        key, subkey = random.split(key)
        W: Array = ig.rvs(size=(size, ), key=key, params={'alpha': 0.5 * nu, 'beta': 0.5 * nu})
        gamma: Array = jnp.zeros_like(mu)
        return super()._rvs(key=subkey, n=size, W=W, mu=mu, gamma=gamma, sigma=sigma)
    
    # stats
    def stats(self, params: dict) -> dict:
        nu, mu, sigma = self._params_to_tuple(params)

        mean: Array = jnp.where(nu > 1, mu, jnp.full_like(mu, jnp.nan))
        scale: Scalar = jnp.where(nu > 2, nu / (nu - 2), jnp.nan)
        cov: Array = scale * sigma
        return {
            "mean": mean, 
            "median": mu, 
            "mode": mu, 
            "cov": cov, 
            "skewness": jnp.zeros_like(mu),}
    
    # fitting
    def _ldmle_inputs(self, d):
        constraints: tuple = (jnp.array([[1e-8]]).T, 
                            jnp.array([[jnp.inf]]).T)
        params0: jnp.ndarray = jnp.exp(random.normal(key=DEFAULT_RANDOM_KEY, shape=(1, )) * 2.5)
        return {'hyperparams': constraints}, params0

    def _reconstruct_ldmle_params(self, params_arr, loc, shape):
        nu: Scalar = params_arr.reshape(())
        scale: Scalar = jnp.where(nu > 2, (nu - 2) / 2, 1.0)
        return nu, loc, scale * shape
    
    def _reconstruct_ldmle_copula_params(self, params, loc, shape):
        nu: Scalar = params.reshape(())
        return nu, loc, shape
        

    
    
        

mvt_student_t = MvtStudentT("Mvt-Student-T")


"""Tests for correlation and shape matrices."""
import pytest
import jax.numpy as jnp
import numpy as np
from jax import jit

from copulax.multivariate import corr, cov, random_correlation, random_covariance


# Helper functions for testing matrix properties
def is_square(matrix):
    """Check if a matrix is square."""
    return matrix.shape[0] == matrix.shape[1]


def is_symmetric(matrix):
    """Check if a matrix is symmetric."""
    return jnp.allclose(matrix, matrix.T, atol=1e-5)


def is_positive_semi_definite(matrix):
    """Check if a matrix is positive semi-definite."""
    eigenvalues = jnp.linalg.eigvalsh(matrix)
    return jnp.all(eigenvalues >= 0)


def is_positive_definite(matrix):
    """Check if a matrix is positive definite."""
    eigenvalues = jnp.linalg.eigvalsh(matrix)
    return jnp.all(eigenvalues > 0)


def has_ones_on_diagonal(matrix, tol=1e-5):
    """Check if a matrix has ones on the diagonal."""
    diag = jnp.diag(matrix)
    return jnp.allclose(diag, jnp.ones_like(diag), atol=tol)


def is_bounded(matrix, tol=1e-5):
    """Check if a matrix is bounded."""
    return jnp.all(jnp.abs(matrix) <= 1 + tol)


# Test methods
CORRELATION_METHODS = [
    "pearson", "spearman", "kendall", "pp_kendall", 
    "rm_pearson", "rm_spearman", "rm_kendall", "rm_pp_kendall",
    "laloux_pearson", "laloux_spearman", "laloux_kendall", "laloux_pp_kendall"
]


# Tests for correlation matrices
@pytest.mark.parametrize("method", CORRELATION_METHODS)
def test_corr_on_correlated_data(correlated_sample, method):
    correlation = corr(correlated_sample, method=method)

    # Check properties
    assert is_square(correlation), f"{method} correlation matrix is not square"
    assert is_symmetric(correlation), f"{method} correlation matrix is not symmetric"
    assert has_ones_on_diagonal(correlation), f"{method} correlation matrix does not have ones on diagonal"
    assert is_positive_semi_definite(correlation), f"{method} correlation matrix is not positive semi-definite"
    assert is_bounded(correlation), f"{method} correlation matrix is not bounded"

    # Checking works with jit
    jit(corr, static_argnames=("method",))(correlated_sample, method)

    
@pytest.mark.parametrize("method", CORRELATION_METHODS)
def test_corr_on_uncorrelated_data(uncorrelated_sample, method):
    correlation = corr(uncorrelated_sample, method=method)
    # Check properties
    assert is_square(correlation), f"{method} correlation matrix is not square with uncorrelated data"
    assert is_symmetric(correlation), f"{method} correlation matrix is not symmetric with uncorrelated data"
    assert has_ones_on_diagonal(correlation), f"{method} correlation matrix does not have ones on diagonal with uncorrelated data"
    assert is_positive_semi_definite(correlation), f"{method} correlation matrix is not positive semi-definite with uncorrelated data"
    assert is_bounded(correlation), f"{method} correlation matrix is not bounded with uncorrelated data"

    # Checking works with jit
    jit(corr, static_argnames=("method",))(uncorrelated_sample, method)


# Tests for covariance matrices
@pytest.mark.parametrize("method", CORRELATION_METHODS)
def test_cov_on_correlated_data(correlated_sample, method):
    covariance = cov(correlated_sample, method=method)
    # Check properties
    assert is_square(covariance), f"{method} covariance matrix is not square"
    assert is_symmetric(covariance), f"{method} covariance matrix is not symmetric"
    assert is_positive_definite(covariance), f"{method} covariance matrix is not positive definite"

    # Checking works with jit
    jit(cov, static_argnames=("method",))(correlated_sample, method)
    
@pytest.mark.parametrize("method", CORRELATION_METHODS)
def test_cov_on_uncorrelated_data(uncorrelated_sample, method):
    covariance = cov(uncorrelated_sample, method=method)
    # Check properties
    assert is_square(covariance), f"{method} covariance matrix is not square with uncorrelated data"
    assert is_symmetric(covariance), f"{method} covariance matrix is not symmetric with uncorrelated data"
    assert is_positive_definite(covariance), f"{method} covariance matrix is not positive definite with uncorrelated data"

    # Checking works with jit
    jit(cov, static_argnames=("method",))(uncorrelated_sample, method)


sizes = tuple((2, 3, 5, 100))

@pytest.mark.parametrize("n", sizes)
def test_random_correlation(n):
    """Test random correlation matrix generation."""
    random_corr = random_correlation(n)
    
    # Check properties
    assert is_square(random_corr), "Random correlation matrix is not square"
    assert random_corr.shape == (n, n), "Random correlation matrix has incorrect shape"
    assert is_symmetric(random_corr), "Random correlation matrix is not symmetric"
    assert has_ones_on_diagonal(random_corr, 1e-5), "Random correlation matrix does not have ones on diagonal"
    assert is_positive_semi_definite(random_corr), "Random correlation matrix is not positive semi-definite"
    assert is_bounded(random_corr), "Random correlation matrix is not bounded"

    # Checking works with jit
    jit(random_correlation, static_argnames=("size",))(n)


@pytest.mark.parametrize("n", sizes)
def test_random_covariance(n):
    """Test random covariance matrix generation."""
    random_vars = np.random.uniform(size=(n, 1))
    random_cov = random_covariance(random_vars)
    
    # Check properties
    assert is_square(random_cov), "Random covariance matrix is not square"
    assert random_cov.shape == (n, n), "Random covariance matrix has incorrect shape"
    assert is_symmetric(random_cov), "Random covariance matrix is not symmetric"
    assert is_positive_definite(random_cov), "Random covariance matrix is not positive definite"

    # Checking works with jit
    jit(random_covariance)(random_vars)


# # Edge case tests
# def test_corr_small_sample(correlated_small_sample):
#     """Test correlation matrices with small samples."""
#     for method in CORRELATION_METHODS:
#         correlation = corr(correlated_small_sample, method=method)
#         assert is_square(correlation), f"{method} correlation matrix is not square with small sample"
#         assert is_symmetric(correlation), f"{method} correlation matrix is not symmetric with small sample"
#         assert has_ones_on_diagonal(correlation), f"{method} correlation matrix does not have ones on diagonal with small sample"
#         assert is_positive_definite(correlation), f"{method} correlation matrix is not positive definite with small sample"


# def test_cov_small_sample(correlated_small_sample):
#     """Test covariance matrices with small samples."""
#     for method in CORRELATION_METHODS:
#         covariance = cov(correlated_small_sample, method=method)
#         assert is_square(covariance), f"{method} covariance matrix is not square with small sample"
#         assert is_symmetric(covariance), f"{method} covariance matrix is not symmetric with small sample"
#         assert is_positive_semi_definite(covariance), f"{method} covariance matrix is not positive semi-definite with small sample"


"""File containing the copulAX implementation of the normal distribution."""
import jax.numpy as jnp
from jax import lax, random
from jax._src.typing import ArrayLike, Array
from jax.scipy import special

from copulax._src._distributions import Univariate
from copulax._src.typing import Scalar
from copulax._src.univariate._utils import _univariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY


class Normal(Univariate):
    r"""The normal / Gaussian distribution is a continuous 'bell shaped' 2 
    parameter family.

    The normal distribution is defined as:

    .. math::

        f(x|\mu, \sigma) = \frac{1}{\sqrt{2\pi\sigma^2}} \exp\left(-\frac{(x - \mu)^2}{2\sigma^2}\right)

    where :math:`\mu` is the mean and :math:`\sigma` the standard deviation 
    of the data.

    https://en.wikipedia.org/wiki/Normal_distribution
    """
    @classmethod
    def _params_dict(cls, mu: Scalar, sigma: Scalar) -> dict:
        d: dict = {"mu": mu, "sigma": sigma}
        return cls._args_transform(d)
    
    def _params_to_tuple(self, params: dict):
        params = self._args_transform(params)
        return params["mu"], params["sigma"]

    def example_params(self, *args, **kwargs):
        r"""Example parameters for the normal distribution.
        
        This is a two parameter family, with the normal / gaussian being 
        defined by its mean and standard deviation.
        """
        return self._params_dict(mu=0.0, sigma=1.0)
    
    @classmethod
    def _support(cls, *args, **kwargs) -> tuple:
        return -jnp.inf, jnp.inf
    
    def logpdf(self, x: ArrayLike, params: dict) -> Array:
        x, xshape = _univariate_input(x)
        mu, sigma = self._params_to_tuple(params)

        const: jnp.ndarray = -0.5 * jnp.log(2 * jnp.pi)
        c: jnp.ndarray = lax.sub(const, jnp.log(sigma)) 
        e: jnp.ndarray = lax.div(lax.pow(lax.sub(x, mu), 2), 2 * lax.pow(sigma, 2))
        logpdf: jnp.ndarray = lax.sub(c, e)
        return logpdf.reshape(xshape)
    
    def logcdf(self, x: ArrayLike, params: dict) -> Array:
        x, xshape = _univariate_input(x)
        mu, sigma = self._params_to_tuple(params)

        z: jnp.ndarray = lax.div(lax.sub(x, mu), sigma)
        logcdf: jnp.ndarray = special.log_ndtr(z)
        return logcdf.reshape(xshape)
    
    def cdf(self, x: ArrayLike, params: dict) -> Array:
        x, xshape = _univariate_input(x)
        mu, sigma = self._params_to_tuple(params)

        z: jnp.ndarray = lax.div(lax.sub(x, mu), sigma)
        cdf: jnp.ndarray = special.ndtr(z)
        return cdf.reshape(xshape)
    
    # ppf
    # def _get_x0(self, params: dict):
    #     return self._args_transform(params)["mu"]

    def _ppf(self, q: ArrayLike, params: dict, *args, **kwargs) -> Array:
        mu, sigma = self._params_to_tuple(params)
        z: jnp.array = jnp.asarray(special.ndtri(q), dtype=float)
        return lax.add(mu, lax.mul(sigma, z))
    
    # sampling
    def rvs(self, size: tuple | Scalar, params: dict, 
            key=DEFAULT_RANDOM_KEY) -> Array:
        mu, sigma = self._params_to_tuple(params)
        return random.normal(key=key, shape=size) * sigma + mu
    
    # stats
    def stats(self, params: dict) -> dict:
        mu, sigma = self._params_to_tuple(params)
        return self._scalar_transform({
            'mean': mu,
            'median': mu,
            'mode': mu,
            'variance': lax.pow(sigma, 2),
            'std': sigma,
            'skewness': 0.0,
            'kurtosis': 0.0,
        })
    
    # fitting
    def fit(self, x: ArrayLike, *args, **kwargs) -> dict:
        x: jnp.ndarray = _univariate_input(x)[0]
        mu: jnp.ndarray = x.mean()
        sigma: jnp.ndarray = x.std()
        return self._params_dict(mu=mu, sigma=sigma)
    

normal = Normal("Normal")


import pytest
import numpy as np

from copulax._src.multivariate._shape import _corr
from copulax.univariate import lognormal, student_t, gamma
from copulax.copulas import gaussian_copula, student_t_copula, gh_copula


# def gen_corr_matrix(num_assets: int) -> np.ndarray:
#     """Generate a correlation matrix for a given number of assets."""
#     # generating a random correlation matrix
#     random_uniform: np.ndarray = np.random.uniform(-1, 1, size=(num_assets, num_assets))
#     lower_triangular: np.ndarray = np.tril(random_uniform)
#     correlation: np.ndarray = lower_triangular + lower_triangular.T
#     np.fill_diagonal(correlation, 1.0)

#     R = _corr._rm_incomplete(correlation, 1e-5)  # 1e-11 too small precision for float32 -> only 6-7 digits of precision
#     eigenvalues = np.linalg.eigvalsh(R)
#     if np.any(eigenvalues <= 0):
#         return gen_corr_matrix(num_assets)
#     return R


# def gen_gaussian_copula_params(num_assets: int) -> dict:
#     """Generate parameters for a Gaussian copula."""
#     mu = np.zeros(num_assets)
#     sigma = gen_corr_matrix(num_assets)
#     return {"copula": {"mu": mu, "sigma": sigma}}


# def gen_student_t_copula_params(num_assets: int) -> dict:
#     """Generate parameters for a Student's t copula."""
#     nu = 3.0 + np.abs(np.random.normal(scale=5))
#     mu = np.zeros(num_assets)
#     sigma = gen_corr_matrix(num_assets)
#     return {"copula": {"nu": nu, "mu": mu, "sigma": sigma}}


# def gen_gh_copula_params(num_assets: int) -> dict:
#     """Generate parameters for a Gaussian copula."""
#     scalars = np.random.normal(size=3)
#     lamb = scalars[0]
#     chi, psi = np.exp(scalars[1:])
#     mu = np.zeros(num_assets)
#     gamma = np.random.normal(size=num_assets, scale = 2.5)
#     sigma = gen_corr_matrix(num_assets)
#     return {"copula": {"lamb": lamb, "chi": chi, "psi": psi, 
#                        "mu": mu, "gamma": gamma, "sigma": sigma}}


# NUM_ASSETS: int = 3
# NUM_SAMPLES: int = 100

# @pytest.fixture(scope='package', autouse=True)
# def continuous_sample():
#     """Generate a multivariate normal sample."""
#     mu = np.random.uniform(-5, 5, NUM_ASSETS)
#     R = gen_corr_matrix(NUM_ASSETS)
#     sigma = np.random.uniform(0.5, 2.5, NUM_ASSETS)
#     sigma_diag = np.diag(sigma)
#     C = sigma_diag @ R @ sigma_diag
#     sample = np.random.multivariate_normal(mu, C, NUM_SAMPLES)

#     # giving data a non-normal profile
#     sample[:, 0] = np.exp(sample[:, 0] / np.abs(sample[:, 0]).max()) # lognormal
#     sample[:, -1] = np.abs(sample[:, -1]) # strictly positive
#     return sample


# @pytest.fixture(scope='package', autouse=True)
# def u_sample():
#     """Generate a sample of uniform marginals."""
#     return np.random.uniform(size=(NUM_SAMPLES, NUM_ASSETS))

# def get_all_args() -> dict:
#     """Generate arguments for copula tests."""
#     # generating marginals
#     marginals = (lognormal, {'mu': 0.0, 'sigma': 1.0}), \
#                 (student_t, {'nu': 3.0, 'mu': 0.0, 'sigma': 1.0}), \
#                 (gamma, {'alpha': 2.0, 'beta': 1.0})
    
#     all_args: dict = {}
#     copula_names = ('gaussian', 'student_t', 'gh')
#     for name in copula_names:
#         # initialising
#         obj = eval(f'{name}_copula')
#         all_args[obj] = {}

#         # generating parameters
#         copula_params = eval(f'gen_{name}_copula_params')(NUM_ASSETS)
#         all_args[obj]['params'] = {'marginals': marginals, **copula_params}
        
#     return all_args
# # TODO: the above has been set up to be an input in a similar way to the shape and multivariate tests. 
# # i,e, with the paramertrize decorator. hence these funcs should be imported into the test file


NUM_SAMPLES: int = 100
NUM_ASSETS: int = 3


np.random.seed(0)

def gen_uncorrelated_data(num_samples: int, num_assets: int):
    return np.random.normal(size=(num_samples, num_assets))


def gen_correlated_data(num_assets: int, num_samples: int):
    # generating a random correlation matrix
    random_uniform: np.ndarray = np.random.uniform(-1, 1, size=(num_assets, num_assets))
    lower_triangular: np.ndarray = np.tril(random_uniform)
    correlation: np.ndarray = lower_triangular + lower_triangular.T
    np.fill_diagonal(correlation, 1.0)

    # converting to covariance matrix
    sigma_diag: np.ndarray = np.diag(np.random.uniform(1, 4, num_assets))
    covariance_matrix: np.ndarray = sigma_diag @ correlation @ sigma_diag

    # generating correlated data
    return np.random.multivariate_normal(np.zeros(num_assets), covariance_matrix, num_samples)


@pytest.fixture(scope='package', autouse=True)
def u_sample():
    """Generate a sample of uniform marginals."""
    eps = 1e-2
    u_raw = np.random.uniform(size=(NUM_SAMPLES, NUM_ASSETS))
    return np.clip(u_raw, eps, 1 - eps)


@pytest.fixture(scope='package', autouse=True)
def datasets():
    return {'uncorrelated_sample': gen_uncorrelated_data(NUM_SAMPLES, NUM_ASSETS),
            'correlated_sample': gen_correlated_data(NUM_ASSETS, NUM_SAMPLES), 
            
            'too_large_dim_sample': gen_uncorrelated_data(NUM_SAMPLES, 100),
            }


"""File containing the copulAX implementation of the multivariate normal 
distribution."""
import jax.numpy as jnp
from jax import lax, random, jit
from jax._src.typing import ArrayLike, Array
from jax.scipy import special

from copulax._src._distributions import Multivariate
from copulax._src.typing import Scalar
from copulax._src.multivariate._utils import _multivariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.multivariate._shape import cov


class MvtNormal(Multivariate):
    r"""The multivariate normal / Gaussian distribution is a 
    generalization of the univariate normal distribution to d > 1 
    dimensions.

    https://en.wikipedia.org/wiki/Multivariate_normal_distribution

    .. math::

        f(x|\mu, \Sigma) = \frac{1}{(2\pi)^{n/2}|\Sigma|^{1/2}} \exp\left(-\frac{1}{2}(x - \mu)^T \Sigma^{-1} (x - \mu)\right)

    where :math:`\mu` is the mean vector and :math:`\sigma` the 
    variance-covariance matrix of the data distribution.
    """
    def _classify_params(self, params: dict) -> dict:
        return super()._classify_params(
            params=params, vector_names=('mu',), shape_names=('sigma',), 
            symmetric_shape_names=('sigma',))

    def _params_dict(self, mu: ArrayLike, sigma: ArrayLike) -> dict:
        d: dict = {"mu": mu, "sigma": sigma}
        return self._args_transform(d)
    
    def _params_to_tuple(self, params: dict) -> tuple:
        params = self._args_transform(params)
        return params["mu"], params["sigma"]
    
    def example_params(self, dim: int = 3, *args, **kwargs) -> dict:
        r"""Example parameters for the multivariate normal distribution.
        
        This is a two parameter family, defined by the mean / location 
        vector `mu` and the variance-covariance matrix `sigma`.

        Args:
            dim: int, number of dimensions of the multivariate normal 
                distribution. Default is 3.
        """
        return self._params_dict(mu=jnp.zeros((dim, 1)), 
                                 sigma=jnp.eye(dim, dim))
    
    def support(self, params: dict) -> Array:
        return super().support(params=params)

    def logpdf(self, x: ArrayLike, params: dict) -> Array:
        x, yshape, n, d = _multivariate_input(x)
        mu, sigma = self._params_to_tuple(params)

        const: jnp.ndarray = -0.5 * (d * jnp.log(2 * jnp.pi) + jnp.linalg.slogdet(sigma)[1])

        sigma_inv: Array = jnp.linalg.inv(sigma)
        Q: jnp.ndarray = self._calc_Q(x=x, mu=mu, sigma_inv=sigma_inv)

        logpdf: jnp.ndarray = -0.5 * Q + const
        return logpdf.reshape(yshape)
    
    # sampling
    def rvs(self, size: int, params: dict, key=DEFAULT_RANDOM_KEY) -> Array:
        mu, sigma = self._params_to_tuple(params)
        return random.multivariate_normal(key=key, mean=mu.flatten(), 
                                          cov=sigma, shape=(size, ))
    
    # stats
    def stats(self, params: dict) -> dict:
        mu, sigma = self._params_to_tuple(params)
        return {
            "mean": mu, 
            "median": mu, 
            "mode": mu, 
            "cov": sigma, 
            "skewness": jnp.zeros_like(mu),}

    # fitting
    def fit(self, x: ArrayLike, sigma_method: str = 'pearson', 
            *args, **kwargs) -> dict:
        r"""Fits the multivariate normal distribution to the data.

        Note:
            If you intend to jit wrap this function, ensure that 
            'sigma_method' is a static argument.

        Args:
            x: arraylike, data to fit the distribution to.
            sigma_method: str, method to estimate the covariance matrix,
                sigma. See copulax.multivariate.cov for available 
                methods.

        Returns:
            dict containing the fitted parameters.
        """
        x, _, _, d = _multivariate_input(x)
        mu: jnp.ndarray = jnp.mean(x, axis=0)
        sigma: jnp.ndarray = cov(x=x, method=sigma_method)
        return self._params_dict(mu=mu, sigma=sigma)
    
    def _fit_copula(self, u, corr_method = 'pearson', *args, **kwargs):
        d: dict = super()._fit_copula(u, corr_method, *args, **kwargs)
        return self._params_dict(mu=d["mu"], sigma=d["sigma"])
        

mvt_normal = MvtNormal("Mvt-Normal")

# TODO: i believe i have finished updating mvt_normal. check if works correctly and if so move onto other mvts


"""File containing the copulAX implementation of the Gamma distribution."""
import jax.numpy as jnp
from jax import lax, random, scipy
from jax._src.typing import ArrayLike, Array
from tensorflow_probability.substrates import jax as tfp

from copulax._src._distributions import Univariate
from copulax._src.typing import Scalar
from copulax._src.univariate._utils import _univariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.optimize import projected_gradient
from copulax._src.univariate.lognormal import lognormal


class Gamma(Univariate):
    r"""The gamma distribution is a two-parameter family of continuous probability
    distributions, which includes the exponential, Erlang and chi-squared 
    distributions as special cases.

    We use the rate parameterization of the gamma distribution specified by 
    McNeil et al (2005).

    https://en.wikipedia.org/wiki/Gamma_distribution"""
    @classmethod
    def _params_dict(cls, alpha: Scalar, beta: Scalar) -> dict:
        d: dict = {"alpha": alpha, "beta": beta}
        return cls._args_transform(d)

    def _params_to_tuple(self, params: dict):
        params = self._args_transform(params)
        return params["alpha"], params["beta"]
    
    def example_params(self, *args, **kwargs):
        r"""Example parameters for the gamma distribution.
        
        This is a two parameter family, defined by alpha and beta 
        parameters. Here we adopt the rate parameterization of the gamma.
        """
        return self._params_dict(alpha=1.0, beta=1.0)
    
    @classmethod
    def _support(cls, *args, **kwargs) -> tuple:
        return 0.0, jnp.inf
    
    def _stable_logpdf(self, stability: Scalar, x: ArrayLike, params: dict) -> Array:
        x, xshape = _univariate_input(x)
        alpha, beta = self._params_to_tuple(params)
        
        logpdf: jnp.ndarray = (alpha * jnp.log(beta + stability) 
                               - lax.lgamma(alpha) 
                               + (alpha - 1) * jnp.log(x) 
                               - beta * x)
        return logpdf.reshape(xshape)

    def logpdf(self, x: ArrayLike, params: dict) -> Array:
        return super().logpdf(x=x, params=params)

    def pdf(self, x: ArrayLike, params: dict) -> Array:
        return super().pdf(x=x, params=params)
    
    def logcdf(self, x: ArrayLike, params: dict) -> Array:
        return super().logcdf(x=x, params=params)
    
    def cdf(self, x: ArrayLike, params: dict) -> Array:
        x, xshape = _univariate_input(x)
        alpha, beta = self._params_to_tuple(params)
        cdf: jnp.ndarray = scipy.special.gammainc(a=alpha, x=beta*x)
        return cdf.reshape(xshape)
    
    # ppf
    # def _get_x0(self, params: dict):
    #     return self.stats(params=params)["mean"]

    def _ppf(self, q: ArrayLike, params: dict, *args, **kwargs) -> Array:
        alpha, beta = self._params_to_tuple(params)
        return tfp.math.igammainv(a=alpha, p=q) / beta
    
    # sampling
    def rvs(self, size: tuple | Scalar, params: dict, key: Array = DEFAULT_RANDOM_KEY) -> Array:
        alpha, beta = self._params_to_tuple(params)
        unscales_rvs: jnp.ndarray = random.gamma(key, shape=size, a=alpha)
        return unscales_rvs / beta
    
    # stats
    def stats(self, params: dict) -> dict:
        alpha, beta = self._params_to_tuple(params)
        mean: float = alpha / beta
        mode: float = jnp.where(alpha >= 1.0, (alpha - 1) / beta, 0.0)
        variance: float = alpha / (beta ** 2)
        std: float = jnp.sqrt(variance)
        skewness: float = 2.0 / jnp.sqrt(alpha)
        kurtosis: float = 6.0 / alpha
        return self._scalar_transform({
            "mean": mean, 
            "mode": mode, 
            "variance": variance, 
            "std": std, 
            "skewness": skewness, 
            "kurtosis": kurtosis})
    
    # fitting
    # def _params_from_array(self, params_arr, *args, **kwargs):
    #     alpha, beta = params_arr
    #     return self._args_transform({})

    def _fit_mle(self, x: ArrayLike, lr: float, maxiter: int) -> dict:
        beta0: float = self.rvs(size=(), params=self.example_params())
        alpha0: float = x.mean() * beta0
        params0: jnp.ndarray = jnp.array([alpha0, beta0])

        res = projected_gradient(f=self._mle_objective, x0=params0, 
                                projection_method='projection_non_negative', 
                                x=x, lr=lr, maxiter=maxiter)
        alpha, beta = res['x']
        return self._params_dict(alpha=alpha, beta=beta)#, res['fun']
    
    def fit(self, x: ArrayLike, lr: float = 0.1, maxiter: int = 100) -> dict:
        r"""Fit the distribution to the input data.
        
        Args:
            x (ArrayLike): The input data to fit the distribution to.
            lr (float): Learning rate for the fitting process.
            maxiter (int): Maximum number of iterations for the fitting process.
        
        Returns:
            dict: The fitted distribution parameters.
        """
        x: jnp.ndarray = _univariate_input(x)[0]
        return self._fit_mle(x=x, lr=lr, maxiter=maxiter)
    

gamma = Gamma("Gamma")


# """Tests for the univariate_fitter function"""
from jax import jit
import pytest
from typing import Callable

from copulax.univariate import univariate_fitter, normal, student_t, lognormal
from copulax._src.typing import Scalar


METRICS = tuple(("aic", "bic", "loglikelihood"))
USER_DEFINED_DISTS = tuple((normal, student_t, lognormal))
DISTS = tuple(("all", "common", "continuous", "discrete", "common continuous", "common discrete", USER_DEFINED_DISTS))
TEST_OPTIONS = tuple((metric, dists) for metric in METRICS for dists in DISTS)


@pytest.mark.parametrize("metric, dists", TEST_OPTIONS)
def test_univariate_fitter(metric, dists, continuous_data):
    # testing univariate fitter works
    best_index, fitted_dists = univariate_fitter(
        x=continuous_data, metric=metric, distributions=dists)
    
    # checking properties
    desc = f"metric = {metric} and dists = {dists}"
    assert isinstance(best_index, Scalar), f"Best index is not a scalar when {desc}."
    assert isinstance(fitted_dists, tuple), f"Fitted distributions are not a tuple when {desc}."
    assert all(isinstance(dist, dict) for dist in fitted_dists), f"Fitted distributions tuple does not contain dictionaries when {desc}."
    assert all(("params" in dist_res and "metric" in dist_res and "dist" in dist_res) for dist_res in fitted_dists), f"Fitted distribution outputs do not contain all of 'params', 'metric' and 'dist' keys when {desc}."

    # testing univariate fitter works when jitted
    jitted_func: Callable = jit(univariate_fitter, static_argnames=['metric', 'distributions'])
    best_index_jit, fitted_dists_jit = jitted_func(
        x=continuous_data, metric=metric, distributions=dists)
    assert isinstance(best_index_jit, Scalar), f"Best index is not a scalar when jitted for {desc}."
    assert isinstance(fitted_dists_jit, tuple), f"Fitted distributions are not a tuple when jitted for {desc}."



"""Helper functions for testing."""
from jax import grad, jit
from jax import numpy as jnp
import numpy as np
import warnings

from copulax._src.typing import Scalar


@jit
def jittable(dist):
    return dist.example_params()


def check_uvt_params(params, s):
    assert isinstance(params, dict), f"{s} params is not a dict"
    assert len(params) > 0, f"{s} is empty"
    assert all(isinstance(k, str) for k in params.keys()), f"{s} params keys are not strings"
    assert all(isinstance(v, Scalar) for v in params.values()), f"{s} params values are not scalars"
    assert all(v.ndim == 0 and v.shape == () and v.size == 1 for v in params.values()), f"{s} params values are not scalars"
    assert any(jnp.isnan(v) for v in params.values()) == False, f"{s} params values are NaN"
    assert all(jnp.isfinite(v) for v in params.values()) == True, f"{s} params values are not finite"


def check_mvt_params(params, s):
    assert isinstance(params, dict), f"{s} params is not a dict"
    assert len(params) > 0, f"{s} is empty"
    assert all(isinstance(k, str) for k in params.keys()), f"{s} params keys are not strings"
    assert all(isinstance(v, jnp.ndarray) for v in params.values()), f"{s} params values are not arrays"
    assert all((v.ndim == 0 and v.shape == () and v.size == 1) or (v.ndim == 2 and v.shape == (v.size, 1) and v.size > 1) or (v.ndim == 2 and v.shape == (int(v.size ** 0.5), int(v.size ** 0.5)) and v.size > 1) for v in params.values()), f"{s} params values are not scalars, 1D-vectors or 2D-square matrices"
    assert any(jnp.any(jnp.isnan(v)) for v in params.values()) == False, f"{s} params values are NaN"
    assert all(jnp.all(jnp.isfinite(v)) for v in params.values()) == True, f"{s} params values are not finite"


def correct_uvt_shape(x, output, dist, method):
    """Check if the shape of the univariate output array is correct"""
    assert isinstance(output, jnp.ndarray), f"{method} output is not a JAX array for {dist}"
    assert output.size == output.size, f"{method} size mismatch for {dist}"
    assert output.shape == output.shape, f"{method} shape mismatch for {dist}"
    assert output.ndim == 1, f"{method} is not 1D for {dist}"


def correct_mvt_shape(x, output, dist, method):
    """Check if the shape of the multivariate output array is correct."""
    assert isinstance(output, jnp.ndarray), f"{method} output is not a JAX array for {dist}"
    expected_shape: tuple = (x.shape[0], 1)
    assert output.shape == expected_shape, f"{method} shape mismatch for {dist}"


def two_dim(output):
    """Check if the input array is two-dimensional."""
    return output.ndim == 2


def is_positive(output):
    """Check if the output array is positive."""
    return np.all(output >= 0)


def no_nans(output):
    """Check if the output array has no NaNs."""
    return not np.any(np.isnan(output))


def is_finite(output):
    """Check if the output array contains only finite values."""
    return np.all(np.isfinite(output))


# def gradients(func, s, data, params, params_error: bool = True, **kwargs):
#     """Calculate the gradients of the output."""
#     new_func = lambda x, p: func(x, params=p, **kwargs).sum()
#     x_grad, params_grad = grad(new_func, argnums=[0, 1])(data, params)
#     params_grad = tuple(params_grad.values())
#     assert no_nans(x_grad), f"{s} gradient contains NaNs for data argument"
#     assert is_finite(x_grad), f"{s} gradient contains non-finite values for data argument"

#     params_nans_res = no_nans(params_grad), f"{s} gradient contains NaNs for params argument"
#     params_finite_res = is_finite(params_grad), f"{s} gradient contains non-finite values for params argument"
#     if params_error:
#         assert params_nans_res[0], params_nans_res[1]
#         assert params_finite_res[0], params_finite_res[1]
#     else:
#         if not params_nans_res[0]:
#             warnings.warn(params_nans_res[1])
#         elif not params_finite_res[0]:
#             warnings.warn(params_finite_res[1])


def gradients(func, s, data, params, params_error: bool = True, **kwargs):
    """Calculate the gradients of the output."""
    new_func = lambda x, p: func(x, params=p, **kwargs).sum()
    x_grad = grad(new_func, argnums=[0,])(data, params)
    # breakpoint()
    # params_grad = tuple(params_grad.values())
    assert no_nans(x_grad), f"{s} gradient contains NaNs for data argument"
    assert is_finite(x_grad), f"{s} gradient contains non-finite values for data argument"

    # params_nans_res = no_nans(params_grad), f"{s} gradient contains NaNs for params argument"
    # params_finite_res = is_finite(params_grad), f"{s} gradient contains non-finite values for params argument"
    # if params_error:
    #     assert params_nans_res[0], params_nans_res[1]
    #     assert params_finite_res[0], params_finite_res[1]
    # else:
    #     if not params_nans_res[0]:
    #         warnings.warn(params_nans_res[1])
    #     elif not params_finite_res[0]:
    #         warnings.warn(params_finite_res[1])


def is_scalar(output):
    """Check if the output is a scalar."""
    return np.asarray(output).flatten().size == 1


def check_metric_output(dist, output, metric_name):
    assert isinstance(output, Scalar) and output.shape == () and output.size == 1, f"{dist} {metric_name} is non-scalar."
    assert no_nans(output), f"{dist} {metric_name} contains NaNs."
    # assert is_finite(output), f"{dist} {metric_name} contains non-finite values."


from jax._src.prng import PRNGKeyArray
import pytest

from copulax import get_random_key, get_api_random_key, get_local_random_key

def _check_valid_key(key, s: str) -> None:
    """Checks if the key is a valid JAX PRNGKey."""
    assert isinstance(key, PRNGKeyArray), f"{s} key is not a valid JAX PRNGKey."


def test_get_api_random_key():
    # testing api method
    try:
        api_key = get_api_random_key()
        _check_valid_key(api_key, "API")
    except ConnectionError as e:
        if str(e) == "Unable to connect to random.org after max = 10 attempts.":
            pytest.skip("Skipping test due to connection error with random.org API.")
        else:
            pytest.fail(f"Failed to get API key: {e}")


def test_wrapper_get_api_random_key():
    try:
        wrapper_api_key = get_random_key(local=False)
        _check_valid_key(wrapper_api_key, "Wrapper API")
    except ConnectionError as e:
        if str(e) == "Unable to connect to random.org after max = 10 attempts.":
            pytest.skip("Skipping test due to connection error with random.org API.")
        else:
            pytest.fail(f"Failed to get API key: {e}")


def test_get_local_random_key():
    # testing local method
    local_key = get_local_random_key()
    _check_valid_key(local_key, "Local")


def test_wrapper_get_local_random_key():
    # testing wrapper local method
    wrapper_local_key = get_random_key(local=True)
    _check_valid_key(wrapper_local_key, "Wrapper Local")



"""File containing utility functions for univariate distributions."""
import jax.numpy as jnp
from jax._src.typing import ArrayLike


def _univariate_input(x: ArrayLike, dtype=float) -> tuple[jnp.ndarray, tuple[int]]:
    """Ensures all input arrays are of the same dtype and (n, 1) shape."""
    x_arr: jnp.array = jnp.asarray(x, dtype=dtype)
    xshape = x_arr.shape
    return x_arr.reshape((x_arr.size, 1)), xshape



"""File containing the copulAX implementation of the multivariate 
skewed-T distribution."""
import jax.numpy as jnp
from jax import lax, random, jit
from jax._src.typing import ArrayLike, Array
from jax.scipy import special

from copulax._src._distributions import NormalMixture
from copulax._src.typing import Scalar
from copulax._src.multivariate._utils import _multivariate_input
from copulax._src._utils import DEFAULT_RANDOM_KEY
from copulax._src.multivariate._shape import cov, _corr
from copulax._src.univariate.ig import ig
from copulax._src.univariate.skewed_t import skewed_t
from copulax.special import kv
from copulax._src.multivariate.mvt_student_t import mvt_student_t


class MvtSkewedT(NormalMixture):
    r"""The multivariate skewed-T distribution is a generalization of 
    the univariate skewed-T distribution to d > 1 dimensions, which 
    itself is a generalization of the student-t distribution which 
    allows for skewness. It can also be expressed as a limiting case of
    the multivariate generalized hyperbolic distribution (GH) when 
    phi -> 0 in addition to lambda = -0.5*chi.

    We use the 4 parameter McNeil et al (2005) specification of the 
    distribution.
    """

    def _classify_params(self, params: dict) -> tuple:
        return super()._classify_params(
            params=params, scalar_names=('nu', ), vector_names=('mu', 'gamma'), 
            shape_names=('sigma',), symmetric_shape_names=('sigma',))

    def _params_dict(self, nu: Scalar, mu: ArrayLike, gamma: ArrayLike, 
                     sigma: ArrayLike) -> dict:
        d: dict = {"nu": nu, "mu": mu, "gamma": gamma, "sigma": sigma}
        return self._args_transform(d)
    
    def _params_to_tuple(self, params: dict) -> tuple:
        params = self._args_transform(params)
        return (params["nu"], params["mu"], 
                params["gamma"], params["sigma"])
    
    def example_params(self, dim: int = 3, *args, **kwargs):
        return self._params_dict(
            nu=4.5, mu=jnp.zeros((dim, 1)), gamma=jnp.zeros((dim, 1)), 
            sigma=jnp.eye(dim, dim))
    
    def support(self, params: dict) -> Array:
        return super().support(params=params)
    
    def _skt_stable_logpdf(self, stability: Scalar, x: ArrayLike,
                       params: dict) -> Array:
        x, yshape, n, d = _multivariate_input(x)
        nu, mu, gamma, sigma = self._params_to_tuple(params)

        sigma_inv: Array = jnp.linalg.inv(sigma)
        Q: Array = self._calc_Q(x=x, mu=mu, sigma_inv=sigma_inv)
        P: Array = sigma_inv @ gamma
        R: Array = gamma.T @ P
        s: Scalar = 0.5 * (nu + d)
        log_det_sigma: Scalar = jnp.linalg.slogdet(sigma)[1]

        log_c: Scalar = ((1-s) * jnp.log(2) 
                         - lax.lgamma(0.5*nu) 
                         - 0.5 * (d*lax.log(nu*jnp.pi + stability) + log_det_sigma))
        
        log_T: Array = (lax.log(kv(s, jnp.sqrt((nu + Q) * R).flatten()))
                        + ((x.T - mu).T @ P).flatten())
        
        log_B: Array = s * (lax.log(1 + Q / (nu + stability)) 
                            - 0.5 * lax.log(stability + (nu + Q) * R)).flatten()
        
        logpdf: Array = (log_c + log_T - log_B)
        return logpdf.reshape(yshape)
    
    def _stable_logpdf(self, stability: Scalar, x: ArrayLike, params: dict) -> Array:
        gamma: Array = jnp.asarray(params["gamma"])
        return lax.cond(jnp.all(gamma == 0), 
                        lambda x: mvt_student_t._stable_logpdf(
                            stability=stability, x=x, params=params),
                        lambda x: self._skt_stable_logpdf(
                            stability=stability, x=x, params=params), x)
    
    # sampling
    def rvs(self, size: int, params: dict, key: ArrayLike=DEFAULT_RANDOM_KEY) -> Array:
        nu, mu, gamma, sigma = self._params_to_tuple(params)

        key, subkey = random.split(key)
        W: Array = ig.rvs(size=(size, ), key=key, params={'alpha': 0.5 * nu, 'beta': 0.5 * nu})
        return super()._rvs(key=subkey, n=size, W=W, mu=mu, gamma=gamma, sigma=sigma)
    
    # stats
    def stats(self, params: dict) -> dict:
        nu, mu, gamma, sigma = self._params_to_tuple(params)
        ig_stats = ig.stats(params={'alpha': 0.5 * nu, 'beta': 0.5 * nu})
        return self._stats(w_stats=ig_stats, mu=mu, gamma=gamma, sigma=sigma)

    # fitting
    def _ldmle_inputs(self, d):
        lc = jnp.full((d + 1,1), -jnp.inf)
        uc = jnp.full((d + 1,1), jnp.inf)
        
        key1, key2 = random.split(DEFAULT_RANDOM_KEY)
        key2, key3 = random.split(key2)
        params0 = jnp.array([lax.exp(random.normal(key1) * 2.5), *random.normal(key3, (d,))]).flatten()
        return {'hyperparams': (lc, uc)}, params0
    
    def _reconstruct_ldmle_params(self, params_arr, loc, shape):
        d: int = loc.size
        nu_: Scalar = lax.dynamic_slice_in_dim(params_arr, 0, 1)
        nu: Scalar = jnp.exp(nu_).flatten()
        gamma: Array = lax.dynamic_slice_in_dim(params_arr, 1, d).reshape((d, 1))
        ig_stats = skewed_t._get_w_stats(nu=nu)

        mu: Array = loc - ig_stats["mean"] * gamma
        sigma_: Array = (shape - ig_stats["variance"] * jnp.outer(gamma, gamma)) / ig_stats["mean"]
        sigma: Array = _corr._rm_incomplete(sigma_, 1e-5)
        return nu, mu, gamma, sigma
    
    def _reconstruct_ldmle_copula_params(self, params_arr, loc, shape):
        d: int = loc.size
        nu_: Scalar = lax.dynamic_slice_in_dim(params_arr, 0, 1)
        nu: Scalar = jnp.exp(nu_).flatten()
        gamma = lax.dynamic_slice_in_dim(params_arr, 1, d).reshape((d, 1))
        return nu, loc, gamma, shape
    

mvt_skewed_t = MvtSkewedT("Mvt-Skewed-T")